"""Manage Prefect Cloud IP Allowlists."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any

import cyclopts
from rich.panel import Panel
from rich.table import Table

import prefect.cli._app as _cli
from prefect.cli._cloud_utils import confirm_logged_in
from prefect.cli._utilities import (
    exit_with_error,
    exit_with_success,
    with_cli_exception_handling,
)

ip_allowlist_app: cyclopts.App = cyclopts.App(
    name="ip-allowlist",
    alias="ip-allowlists",
    help="Manage Prefect Cloud IP Allowlists.",
    version_flags=[],
    help_flags=["--help"],
)


async def _check_ip_allowlist_access() -> bool:
    """Check if the account has access to IP allowlisting.

    Returns the enforce_ip_allowlist setting value.
    Exits with an error if the account does not have access.
    """
    from prefect.client.cloud import get_cloud_client

    confirm_logged_in(console=_cli.console)

    async with get_cloud_client(infer_cloud_url=True) as client:
        account_settings = await client.read_account_settings()

    if "enforce_ip_allowlist" not in account_settings:
        exit_with_error("IP allowlisting is not available for this account.")

    return account_settings.get("enforce_ip_allowlist", False)


def _print_ip_allowlist_table(ip_allowlist: Any, enabled: bool) -> None:
    from prefect.client.schemas.objects import IPAllowlist

    if TYPE_CHECKING:
        assert isinstance(ip_allowlist, IPAllowlist)

    if not ip_allowlist.entries:
        _cli.console.print(
            Panel(
                "IP allowlist is empty. Add an entry to secure access to your Prefect Cloud account.",
                expand=False,
            )
        )
        return

    red_asterisk_if_not_enabled = "[red]*[/red]" if enabled is False else ""

    table = Table(
        title="IP Allowlist " + red_asterisk_if_not_enabled,
        caption=f"{red_asterisk_if_not_enabled} Enforcement is "
        f"[bold]{'ENABLED' if enabled else '[red]DISABLED[/red]'}[/bold].",
        caption_style="not dim",
    )

    table.add_column("IP Address", style="cyan", no_wrap=True)
    table.add_column("Description", style="blue", no_wrap=False)
    table.add_column("Enabled", style="green", justify="right", no_wrap=True)
    table.add_column("Last Seen", style="magenta", justify="right", no_wrap=True)

    for entry in ip_allowlist.entries:
        table.add_row(
            str(entry.ip_network),
            entry.description,
            str(entry.enabled),
            entry.last_seen or "Never",
            style="dim" if not entry.enabled else None,
        )

    _cli.console.print(table)


def _handle_update_error(error: Any) -> None:
    from prefect.exceptions import PrefectHTTPStatusError

    if TYPE_CHECKING:
        assert isinstance(error, PrefectHTTPStatusError)
    if error.response.status_code == 422 and (
        details := (
            error.response.json().get("detail")
            or error.response.json().get("exception_detail")
        )
    ):
        exit_with_error(f"Error updating allowlist: {details}")
    else:
        raise error


def _parse_ip_network(val: str) -> Any:
    """Parse and validate an IP network argument."""
    from pydantic import BaseModel, IPvAnyNetwork, ValidationError

    class IPNetworkArg(BaseModel):
        raw: str
        parsed: IPvAnyNetwork

    try:
        return IPNetworkArg(raw=val, parsed=val)
    except ValidationError:
        exit_with_error(
            f"Invalid value for 'IP address or range': {val!r} is not a valid"
            " IP address or CIDR range."
        )


@ip_allowlist_app.command(name="enable")
@with_cli_exception_handling
async def ip_allowlist_enable():
    """Enable the IP allowlist for your account. When enabled, if the allowlist is non-empty, then access to your Prefect Cloud account will be restricted to only those IP addresses on the allowlist."""
    from prefect.cli._prompts import confirm
    from prefect.client.cloud import get_cloud_client

    enforcing_ip_allowlist = await _check_ip_allowlist_access()
    if enforcing_ip_allowlist:
        exit_with_success("IP allowlist is already enabled.")

    async with get_cloud_client(infer_cloud_url=True) as client:
        my_access_if_enabled = await client.check_ip_allowlist_access()
        from prefect.logging import get_logger

        get_logger(__name__).debug(my_access_if_enabled.detail)
        if not my_access_if_enabled.allowed:
            exit_with_error(
                f"Error enabling IP allowlist: {my_access_if_enabled.detail}"
            )

        if not confirm(
            "Enabling the IP allowlist will restrict Prefect Cloud API and UI access to only the IP addresses on the list. "
            "Continue?",
            console=_cli.console,
        ):
            exit_with_error("Aborted.")
        await client.update_account_settings({"enforce_ip_allowlist": True})

    exit_with_success("IP allowlist enabled.")


@ip_allowlist_app.command(name="disable")
@with_cli_exception_handling
async def ip_allowlist_disable():
    """Disable the IP allowlist for your account. When disabled, all IP addresses will be allowed to access your Prefect Cloud account."""
    from prefect.client.cloud import get_cloud_client

    await _check_ip_allowlist_access()

    async with get_cloud_client(infer_cloud_url=True) as client:
        await client.update_account_settings({"enforce_ip_allowlist": False})

    exit_with_success("IP allowlist disabled.")


@ip_allowlist_app.command(name="ls")
@with_cli_exception_handling
async def ip_allowlist_ls():
    """Fetch and list all IP allowlist entries in your account."""
    from prefect.client.cloud import get_cloud_client

    enforcing_ip_allowlist = await _check_ip_allowlist_access()

    async with get_cloud_client(infer_cloud_url=True) as client:
        ip_allowlist = await client.read_account_ip_allowlist()
        _print_ip_allowlist_table(ip_allowlist, enabled=enforcing_ip_allowlist)


@ip_allowlist_app.command(name="add")
@with_cli_exception_handling
async def ip_allowlist_add(
    ip_address_or_range: Annotated[
        str,
        cyclopts.Parameter(
            help="An IP address or range in CIDR notation. E.g. 192.168.1.0 or 192.168.1.0/24",
        ),
    ],
    *,
    description: Annotated[
        str | None,
        cyclopts.Parameter(
            "--description",
            alias="-d",
            help="A short description to annotate the entry with.",
        ),
    ] = None,
):
    """Add a new IP entry to your account IP allowlist."""
    from prefect.cli._prompts import confirm
    from prefect.client.cloud import get_cloud_client
    from prefect.client.schemas.objects import IPAllowlistEntry
    from prefect.exceptions import PrefectHTTPStatusError

    ip_arg = _parse_ip_network(ip_address_or_range)
    enforcing_ip_allowlist = await _check_ip_allowlist_access()

    new_entry = IPAllowlistEntry(
        ip_network=ip_arg.parsed, description=description, enabled=True
    )

    async with get_cloud_client(infer_cloud_url=True) as client:
        ip_allowlist = await client.read_account_ip_allowlist()

        existing_entry_with_same_ip = None
        for entry in ip_allowlist.entries:
            if entry.ip_network == ip_arg.parsed:
                existing_entry_with_same_ip = entry
                break

        if existing_entry_with_same_ip:
            if not confirm(
                f"There's already an entry for this IP ({ip_arg.raw}). Do you want to overwrite it?",
                console=_cli.console,
            ):
                exit_with_error("Aborted.")
            ip_allowlist.entries.remove(existing_entry_with_same_ip)

        ip_allowlist.entries.append(new_entry)

        try:
            await client.update_account_ip_allowlist(ip_allowlist)
        except PrefectHTTPStatusError as exc:
            _handle_update_error(exc)

        updated_ip_allowlist = await client.read_account_ip_allowlist()
        _print_ip_allowlist_table(updated_ip_allowlist, enabled=enforcing_ip_allowlist)


@ip_allowlist_app.command(name="remove")
@with_cli_exception_handling
async def ip_allowlist_remove(
    ip_address_or_range: Annotated[
        str,
        cyclopts.Parameter(
            help="An IP address or range in CIDR notation. E.g. 192.168.1.0 or 192.168.1.0/24",
        ),
    ],
):
    """Remove an IP entry from your account IP allowlist."""
    from prefect.client.cloud import get_cloud_client
    from prefect.exceptions import PrefectHTTPStatusError

    ip_arg = _parse_ip_network(ip_address_or_range)
    enforcing_ip_allowlist = await _check_ip_allowlist_access()

    async with get_cloud_client(infer_cloud_url=True) as client:
        ip_allowlist = await client.read_account_ip_allowlist()
        ip_allowlist.entries = [
            entry for entry in ip_allowlist.entries if entry.ip_network != ip_arg.parsed
        ]

        try:
            await client.update_account_ip_allowlist(ip_allowlist)
        except PrefectHTTPStatusError as exc:
            _handle_update_error(exc)

        updated_ip_allowlist = await client.read_account_ip_allowlist()
        _print_ip_allowlist_table(updated_ip_allowlist, enabled=enforcing_ip_allowlist)


@ip_allowlist_app.command(name="toggle")
@with_cli_exception_handling
async def ip_allowlist_toggle(
    ip_address_or_range: Annotated[
        str,
        cyclopts.Parameter(
            help="An IP address or range in CIDR notation. E.g. 192.168.1.0 or 192.168.1.0/24",
        ),
    ],
):
    """Toggle the enabled status of an individual IP entry in your account IP allowlist."""
    from prefect.client.cloud import get_cloud_client
    from prefect.exceptions import PrefectHTTPStatusError

    ip_arg = _parse_ip_network(ip_address_or_range)
    enforcing_ip_allowlist = await _check_ip_allowlist_access()

    async with get_cloud_client(infer_cloud_url=True) as client:
        ip_allowlist = await client.read_account_ip_allowlist()

        found_matching_entry = False
        for entry in ip_allowlist.entries:
            if entry.ip_network == ip_arg.parsed:
                entry.enabled = not entry.enabled
                found_matching_entry = True
                break

        if not found_matching_entry:
            exit_with_error(f"No entry found with IP address `{ip_arg.raw}`.")

        try:
            await client.update_account_ip_allowlist(ip_allowlist)
        except PrefectHTTPStatusError as exc:
            _handle_update_error(exc)

        updated_ip_allowlist = await client.read_account_ip_allowlist()
        _print_ip_allowlist_table(updated_ip_allowlist, enabled=enforcing_ip_allowlist)



"""
Worker command — native cyclopts implementation.

Start and interact with workers.
"""

import asyncio
import json
import os
from enum import Enum
from pathlib import Path
from typing import Annotated, Optional

import cyclopts

import prefect.cli._app as _cli
from prefect.cli._utilities import (
    exit_with_error,
    with_cli_exception_handling,
)

worker_app: cyclopts.App = cyclopts.App(
    name="worker", help="Start and interact with workers."
)


class InstallPolicy(str, Enum):
    ALWAYS = "always"
    IF_NOT_PRESENT = "if-not-present"
    NEVER = "never"
    PROMPT = "prompt"


@worker_app.command()
@with_cli_exception_handling
async def start(
    *,
    worker_name: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--name", alias="-n", help="The name to give to the started worker."
        ),
    ] = None,
    work_pool_name: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--pool", alias="-p", help="The work pool the started worker should poll."
        ),
    ] = None,
    work_queues: Annotated[
        Optional[list[str]],
        cyclopts.Parameter(
            "--work-queue",
            alias="-q",
            help="Work queue names to pull from (repeatable).",
        ),
    ] = None,
    worker_type: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--type",
            alias="-t",
            help="The type of worker to start.",
        ),
    ] = None,
    prefetch_seconds: Annotated[
        Optional[int],
        cyclopts.Parameter(
            "--prefetch-seconds",
            help="Seconds to look ahead for scheduled flow runs. [from PREFECT_WORKER_PREFETCH_SECONDS]",
        ),
    ] = None,
    run_once: Annotated[
        bool,
        cyclopts.Parameter("--run-once", help="Only run worker polling once."),
    ] = False,
    limit: Annotated[
        Optional[int],
        cyclopts.Parameter(
            "--limit",
            alias="-l",
            help="Maximum concurrent flow runs.",
        ),
    ] = None,
    with_healthcheck: Annotated[
        bool,
        cyclopts.Parameter("--with-healthcheck", help="Start a healthcheck server."),
    ] = False,
    install_policy: Annotated[
        InstallPolicy,
        cyclopts.Parameter(
            "--install-policy",
            help="Install policy for worker packages.",
        ),
    ] = InstallPolicy.PROMPT,
    base_job_template: Annotated[
        Optional[Path],
        cyclopts.Parameter(
            "--base-job-template",
            help="Path to JSON file containing base job template.",
        ),
    ] = None,
):
    """Start a worker process to poll a work pool for flow runs."""
    from prefect.cli._prompts import confirm
    from prefect.cli._worker_utils import (
        _check_work_pool_paused,
        _check_work_queues_paused,
        _find_package_for_worker_type,
        _install_package,
        _load_worker_class,
        _retrieve_worker_type_from_pool,
    )
    from prefect.settings import (
        PREFECT_WORKER_HEARTBEAT_SECONDS,
        PREFECT_WORKER_PREFETCH_SECONDS,
    )
    from prefect.utilities.processutils import setup_signal_handlers_worker

    # Prompt for work pool name if not provided (matches typer's prompt=True)
    if work_pool_name is None:
        if _cli.is_interactive():
            from rich.prompt import Prompt

            work_pool_name = Prompt.ask("Work pool name", console=_cli.console)
        else:
            exit_with_error("Missing required option '--pool' / '-p'.")

    # Resolve settings-backed defaults
    if prefetch_seconds is None:
        prefetch_seconds = PREFECT_WORKER_PREFETCH_SECONDS.value()

    is_paused = await _check_work_pool_paused(work_pool_name)
    if is_paused:
        _cli.console.print(
            (
                f"The work pool {work_pool_name!r} is currently paused. This worker"
                " will not execute any flow runs until the work pool is unpaused."
            ),
            style="yellow",
        )

    is_queues_paused = await _check_work_queues_paused(work_pool_name, work_queues)
    if is_queues_paused:
        queue_scope = (
            "All work queues" if not work_queues else "Specified work queue(s)"
        )
        _cli.console.print(
            (
                f"{queue_scope} in the work pool {work_pool_name!r} are currently"
                " paused. This worker will not execute any flow runs until the work"
                " queues are unpaused."
            ),
            style="yellow",
        )

    # Resolve worker type
    if worker_type is None:
        worker_type = await _retrieve_worker_type_from_pool(
            _cli.console, exit_with_error, work_pool_name
        )

    if worker_type == "prefect-agent":
        exit_with_error(
            "'prefect-agent' typed work pools work with Prefect Agents instead of"
            " Workers. Please use the 'prefect agent start' to start a Prefect Agent."
        )

    # Load or install worker class (matches typer's _get_worker_class flow)
    if install_policy == InstallPolicy.ALWAYS:
        package = await _find_package_for_worker_type(_cli.console, worker_type)
        if package:
            await _install_package(_cli.console, package, upgrade=True)

    worker_cls = _load_worker_class(worker_type)

    if worker_cls is None:
        package = await _find_package_for_worker_type(_cli.console, worker_type)
        if package:
            should_install = False
            if install_policy == InstallPolicy.IF_NOT_PRESENT:
                should_install = True
            elif install_policy == InstallPolicy.PROMPT and _cli.is_interactive():
                message = (
                    "Could not find the Prefect integration library for the"
                    f" {worker_type} worker in the current environment."
                    " Install the library now?"
                )
                should_install = confirm(message, default=True)

            if should_install:
                await _install_package(_cli.console, package)
                worker_cls = _load_worker_class(worker_type)

    if worker_cls is None:
        exit_with_error(
            "Unable to start worker. Please ensure you have the necessary dependencies"
            " installed to run your desired worker type."
        )

    worker_process_id = os.getpid()
    setup_signal_handlers_worker(
        worker_process_id, f"the {worker_type} worker", _cli.console.print
    )

    template_contents = None
    if base_job_template is not None:
        template_contents = json.loads(base_job_template.read_text())

    worker = worker_cls(
        name=worker_name,
        work_pool_name=work_pool_name,
        work_queues=work_queues,
        limit=limit,
        prefetch_seconds=prefetch_seconds,
        heartbeat_interval_seconds=int(PREFECT_WORKER_HEARTBEAT_SECONDS.value()),
        base_job_template=template_contents,
    )
    try:
        await worker.start(
            run_once=run_once,
            with_healthcheck=with_healthcheck,
            printer=_cli.console.print,
        )
    except asyncio.CancelledError:
        _cli.console.print(f"Worker {worker.name!r} stopped!", style="yellow")



"""
Deploy command — native cyclopts implementation.

Reuses all business logic from prefect.cli.deploy.* modules, threading
`console` and `is_interactive` as parameters.
"""

from __future__ import annotations

from pathlib import Path
from typing import Annotated, Any

import cyclopts
from rich.prompt import Prompt
from rich.table import Table

import prefect
import prefect.cli._app as _cli
from prefect.cli._utilities import (
    exit_with_error,
    with_cli_exception_handling,
)
from prefect.client.schemas.objects import ConcurrencyLimitConfig

deploy_app: cyclopts.App = cyclopts.App(
    name="deploy",
    help="Create and manage deployments.",
    version_flags=[],
    help_flags=["--help"],
)


@deploy_app.command(name="init")
@with_cli_exception_handling
async def init(
    *,
    name: Annotated[
        str | None,
        cyclopts.Parameter("--name", help="The name to give the project."),
    ] = None,
    recipe: Annotated[
        str | None,
        cyclopts.Parameter("--recipe", help="The recipe to use for the project."),
    ] = None,
    fields: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            "--field",
            alias="-f",
            help=(
                "One or more fields to pass to the recipe (e.g., image_name) in the"
                " format of key=value."
            ),
        ),
    ] = None,
):
    """Initialize a Prefect project."""
    import yaml

    from prefect.deployments import initialize_project

    inputs: dict[str, Any] = {}
    fields = fields or []
    recipe_paths = prefect.__module_path__ / "deployments" / "recipes"

    for field in fields:
        key, value = field.split("=")
        inputs[key] = value

    from prefect.cli._prompts import prompt_select_from_table

    if not recipe and _cli.is_interactive():
        recipes: list[dict[str, Any]] = []
        for r in recipe_paths.iterdir():
            if r.is_dir() and (r / "prefect.yaml").exists():
                with open(r / "prefect.yaml") as f:
                    recipe_data = yaml.safe_load(f)
                    recipe_name = r.name
                    recipe_description = recipe_data.get(
                        "description", "(no description available)"
                    )
                    recipes.append(
                        {"name": recipe_name, "description": recipe_description}
                    )

        selected_recipe = prompt_select_from_table(
            _cli.console,
            "Would you like to initialize your deployment configuration with a recipe?",
            columns=[
                {"header": "Name", "key": "name"},
                {"header": "Description", "key": "description"},
            ],
            data=recipes,
            opt_out_message="No, I'll use the default deployment configuration.",
            opt_out_response={},
        )
        if selected_recipe != {}:
            recipe = selected_recipe["name"]

    if recipe and (recipe_paths / recipe / "prefect.yaml").exists():
        with open(recipe_paths / recipe / "prefect.yaml") as f:
            recipe_inputs = yaml.safe_load(f).get("required_inputs") or {}
        if recipe_inputs:
            if set(recipe_inputs.keys()) < set(inputs.keys()):
                _cli.console.print(
                    (
                        f"Warning: extra fields provided for {recipe!r} recipe:"
                        f" '{', '.join(set(inputs.keys()) - set(recipe_inputs.keys()))}'"
                    ),
                    style="red",
                )
            elif set(recipe_inputs.keys()) > set(inputs.keys()):
                table = Table(
                    title=f"[red]Required inputs for {recipe!r} recipe[/red]",
                )
                table.add_column("Field Name", style="green", no_wrap=True)
                table.add_column(
                    "Description", justify="left", style="white", no_wrap=False
                )
                for field_name, description in recipe_inputs.items():
                    if field_name not in inputs:
                        table.add_row(field_name, description)
                _cli.console.print(table)
                for key, description in recipe_inputs.items():
                    if key not in inputs:
                        inputs[key] = Prompt.ask(key, console=_cli.console)
            _cli.console.print("-" * 15)

    try:
        files = [
            f"[green]{fname}[/green]"
            for fname in initialize_project(name=name, recipe=recipe, inputs=inputs)
        ]
    except ValueError as exc:
        if "Unknown recipe" in str(exc):
            exit_with_error(
                f"Unknown recipe {recipe!r} provided - run"
                " [yellow]`prefect init`[/yellow] to see all available recipes."
            )
        else:
            raise

    files_str = "\n".join(files)
    empty_msg = (
        f"Created project in [green]{Path('.').resolve()}[/green];"
        " no new files created."
    )
    file_msg = (
        f"Created project in [green]{Path('.').resolve()}[/green]"
        f" with the following new files:\n{files_str}"
    )
    _cli.console.print(file_msg if files_str else empty_msg)


@deploy_app.default
@with_cli_exception_handling
async def deploy(
    entrypoint: Annotated[
        str | None,
        cyclopts.Parameter(
            show=False,
            help=(
                "The path to a flow entrypoint within a project, in the form of"
                " `./path/to/file.py:flow_func_name`"
            ),
        ),
    ] = None,
    *,
    names: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            "--name",
            alias="-n",
            help=(
                "The name to give the deployment. Can be a pattern. Examples:"
                " 'my-deployment', 'my-flow/my-deployment', 'my-deployment-*',"
                " '*-flow-name/deployment*'"
            ),
        ),
    ] = None,
    description: Annotated[
        str | None,
        cyclopts.Parameter(
            "--description",
            alias="-d",
            help=(
                "The description to give the deployment. If not provided, the"
                " description will be populated from the flow's description."
            ),
        ),
    ] = None,
    version_type: Annotated[
        str | None,
        cyclopts.Parameter(
            "--version-type",
            help="The type of version to use for this deployment.",
        ),
    ] = None,
    version: Annotated[
        str | None,
        cyclopts.Parameter("--version", help="A version to give the deployment."),
    ] = None,
    tags: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            "--tag",
            alias="-t",
            help=(
                "One or more optional tags to apply to the deployment. Note: tags are"
                " used only for organizational purposes. For delegating work to"
                " workers, use the --work-queue flag."
            ),
        ),
    ] = None,
    concurrency_limit: Annotated[
        int | None,
        cyclopts.Parameter(
            "--concurrency-limit",
            help="The maximum number of concurrent runs for this deployment.",
        ),
    ] = None,
    concurrency_limit_collision_strategy: Annotated[
        str | None,
        cyclopts.Parameter(
            "--collision-strategy",
            help=(
                "Configure the behavior for runs once the concurrency limit is"
                " reached. Falls back to `ENQUEUE` if unset."
            ),
        ),
    ] = None,
    work_pool_name: Annotated[
        str | None,
        cyclopts.Parameter(
            "--pool",
            alias="-p",
            help="The work pool that will handle this deployment's runs.",
        ),
    ] = None,
    work_queue_name: Annotated[
        str | None,
        cyclopts.Parameter(
            "--work-queue",
            alias="-q",
            help=(
                "The work queue that will handle this deployment's runs. It will be"
                " created if it doesn't already exist. Defaults to `None`."
            ),
        ),
    ] = None,
    job_variables: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            "--job-variable",
            json_list=False,
            help=(
                "One or more job variable overrides for the work pool provided in the"
                " format of key=value string or a JSON object"
            ),
        ),
    ] = None,
    cron: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            "--cron",
            help=(
                "A cron string that will be used to set a CronSchedule on the"
                " deployment."
            ),
        ),
    ] = None,
    interval: Annotated[
        list[int] | None,
        cyclopts.Parameter(
            "--interval",
            help=(
                "An integer specifying an interval (in seconds) that will be used to"
                " set an IntervalSchedule on the deployment."
            ),
        ),
    ] = None,
    interval_anchor: Annotated[
        str | None,
        cyclopts.Parameter(
            "--anchor-date",
            help="The anchor date for all interval schedules",
        ),
    ] = None,
    rrule: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            "--rrule",
            help=(
                "An RRule that will be used to set an RRuleSchedule on the deployment."
            ),
        ),
    ] = None,
    timezone: Annotated[
        str | None,
        cyclopts.Parameter(
            "--timezone",
            help="Deployment schedule timezone string e.g. 'America/New_York'",
        ),
    ] = None,
    trigger: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            "--trigger",
            json_list=False,
            help=(
                "Specifies a trigger for the deployment. The value can be a json"
                " string or path to `.yaml`/`.json` file. This flag can be used"
                " multiple times."
            ),
        ),
    ] = None,
    param: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            "--param",
            help=(
                "An optional parameter override, values are parsed as JSON strings"
                " e.g. --param question=ultimate --param answer=42"
            ),
        ),
    ] = None,
    params: Annotated[
        str | None,
        cyclopts.Parameter(
            "--params",
            help=(
                "An optional parameter override in a JSON string format e.g."
                ' --params=\'{"question": "ultimate", "answer": 42}\''
            ),
        ),
    ] = None,
    enforce_parameter_schema: Annotated[
        bool,
        cyclopts.Parameter(
            "--enforce-parameter-schema",
            help=(
                "Whether to enforce the parameter schema on this deployment. If set to"
                " True, any parameters passed to this deployment must match the"
                " signature of the flow."
            ),
        ),
    ] = True,
    deploy_all: Annotated[
        bool,
        cyclopts.Parameter(
            "--all",
            help=(
                "Deploy all flows in the project. If a flow name or entrypoint is also"
                " provided, this flag will be ignored."
            ),
        ),
    ] = False,
    prefect_file: Annotated[
        Path,
        cyclopts.Parameter(
            "--prefect-file",
            help="Specify a custom path to a prefect.yaml file",
        ),
    ] = Path("prefect.yaml"),
    sla: Annotated[
        list[str] | None,
        cyclopts.Parameter(
            "--sla",
            json_list=False,
            help=(
                "Experimental: One or more SLA configurations for the deployment. May"
                " be removed or modified at any time. Currently only supported on"
                " Prefect Cloud."
            ),
        ),
    ] = None,
):
    """Create and update deployments."""
    from prefect.cli.deploy._config import (
        _load_deploy_configs_and_actions,
        _parse_name_from_pattern,
        _pick_deploy_configs,
    )
    from prefect.cli.deploy._core import _run_multi_deploy, _run_single_deploy
    from prefect.settings import get_current_settings

    # Resolve default work pool from settings (typer uses a lambda callback).
    if work_pool_name is None:
        work_pool_name = get_current_settings().deployments.default_work_pool_name

    if job_variables is None:
        job_variables = list()

    concurrency_limit_config = (
        None
        if concurrency_limit is None
        else (
            concurrency_limit
            if concurrency_limit_collision_strategy is None
            else ConcurrencyLimitConfig(
                limit=concurrency_limit,
                collision_strategy=concurrency_limit_collision_strategy,
            ).model_dump()
        )
    )

    options: dict[str, Any] = {
        "entrypoint": entrypoint,
        "description": description,
        "version_type": version_type,
        "version": version,
        "tags": tags,
        "concurrency_limit": concurrency_limit_config,
        "work_pool_name": work_pool_name,
        "work_queue_name": work_queue_name,
        "variables": job_variables,
        "cron": cron,
        "interval": interval,
        "anchor_date": interval_anchor,
        "rrule": rrule,
        "timezone": timezone,
        "triggers": trigger,
        "param": param,
        "params": params,
        "sla": sla,
    }

    try:
        all_deploy_configs, actions = _load_deploy_configs_and_actions(
            prefect_file=prefect_file, console=_cli.console
        )
        parsed_names: list[str] = []
        for name in names or []:
            if "*" in name:
                parsed_names.extend(_parse_name_from_pattern(all_deploy_configs, name))
            else:
                parsed_names.append(name)
        deploy_configs = _pick_deploy_configs(
            all_deploy_configs,
            parsed_names,
            deploy_all,
            console=_cli.console,
            is_interactive=_cli.is_interactive,
        )

        if len(deploy_configs) > 1:
            if any(options.values()):
                _cli.console.print(
                    (
                        "You have passed options to the deploy command, but you are"
                        " creating or updating multiple deployments. These options"
                        " will be ignored."
                    ),
                    style="yellow",
                )
            await _run_multi_deploy(
                deploy_configs=deploy_configs,
                actions=actions,
                deploy_all=deploy_all,
                prefect_file=prefect_file,
                console=_cli.console,
                is_interactive=_cli.is_interactive,
            )
        else:
            deploy_config = deploy_configs[0] if deploy_configs else {}
            options["names"] = [
                name.split("/", 1)[-1] if "/" in name else name for name in parsed_names
            ]
            if not enforce_parameter_schema:
                options["enforce_parameter_schema"] = False
            await _run_single_deploy(
                deploy_config=deploy_config,
                actions=actions,
                options=options,
                prefect_file=prefect_file,
                console=_cli.console,
                is_interactive=_cli.is_interactive,
            )
    except ValueError as exc:
        exit_with_error(str(exc))



from typing import ClassVar

from pydantic import Field
from pydantic_settings import SettingsConfigDict

from prefect.settings.base import PrefectBaseSettings, build_settings_config


class WorkerWebserverSettings(PrefectBaseSettings):
    model_config: ClassVar[SettingsConfigDict] = build_settings_config(
        ("worker", "webserver")
    )

    host: str = Field(
        default="0.0.0.0",
        description="The host address the worker's webserver should bind to.",
    )

    port: int = Field(
        default=8080,
        description="The port the worker's webserver should bind to.",
    )


class WorkerSettings(PrefectBaseSettings):
    model_config: ClassVar[SettingsConfigDict] = build_settings_config(("worker",))

    debug_mode: bool = Field(
        default=False,
        description=(
            "If True, enables debug mode for the worker only. "
            "Unlike PREFECT_DEBUG_MODE, this setting does not propagate "
            "to flow runs executed by the worker."
        ),
    )

    heartbeat_seconds: float = Field(
        default=30,
        description="Number of seconds a worker should wait between sending a heartbeat.",
    )

    query_seconds: float = Field(
        default=10,
        description="Number of seconds a worker should wait between queries for scheduled work.",
    )

    prefetch_seconds: float = Field(
        default=10,
        description="The number of seconds into the future a worker should query for scheduled work.",
    )

    enable_cancellation: bool = Field(
        default=False,
        description=(
            "Enable worker-side flow run cancellation for pending flow runs. "
            "When enabled, the worker will terminate infrastructure for flow runs "
            "that are cancelled while still in PENDING state (before the runner starts)."
        ),
    )

    cancellation_poll_seconds: float = Field(
        default=120,
        description=(
            "Number of seconds between polls for cancelling flow runs. "
            "Used as a fallback when the WebSocket connection for real-time "
            "cancellation events is unavailable."
        ),
    )

    webserver: WorkerWebserverSettings = Field(
        default_factory=WorkerWebserverSettings,
        description="Settings for a worker's webserver",
    )



from __future__ import annotations

from contextlib import contextmanager
from typing import Generator, Literal, Optional
from uuid import UUID

from prefect.client.schemas.objects import ConcurrencyLeaseHolder
from prefect.client.schemas.responses import (
    ConcurrencyLimitWithLeaseResponse,
    MinimalConcurrencyLimitResponse,
)
from prefect.concurrency._asyncio import (
    aacquire_concurrency_slots,
    aacquire_concurrency_slots_with_lease,
    arelease_concurrency_slots_with_lease,
)
from prefect.concurrency._events import (
    emit_concurrency_acquisition_events,
    emit_concurrency_release_events,
)
from prefect.concurrency._leases import maintain_concurrency_lease
from prefect.utilities.asyncutils import run_coro_as_sync


def release_concurrency_slots_with_lease(lease_id: UUID) -> None:
    run_coro_as_sync(arelease_concurrency_slots_with_lease(lease_id))


def acquire_concurrency_slots(
    names: list[str],
    slots: int,
    mode: Literal["concurrency", "rate_limit"] = "concurrency",
    timeout_seconds: Optional[float] = None,
    max_retries: Optional[int] = None,
    strict: bool = False,
) -> list[MinimalConcurrencyLimitResponse]:
    result = run_coro_as_sync(
        aacquire_concurrency_slots(
            names, slots, mode, timeout_seconds, max_retries, strict
        )
    )
    return result


def acquire_concurrency_slots_with_lease(
    names: list[str],
    slots: int,
    mode: Literal["concurrency", "rate_limit"] = "concurrency",
    timeout_seconds: Optional[float] = None,
    max_retries: Optional[int] = None,
    lease_duration: float = 300,
    strict: bool = False,
    holder: "Optional[ConcurrencyLeaseHolder]" = None,
    suppress_warnings: bool = False,
) -> ConcurrencyLimitWithLeaseResponse:
    result = run_coro_as_sync(
        aacquire_concurrency_slots_with_lease(
            names,
            slots,
            mode,
            timeout_seconds,
            max_retries,
            lease_duration,
            strict,
            holder,
            suppress_warnings,
        )
    )
    return result


@contextmanager
def concurrency(
    names: str | list[str],
    occupy: int = 1,
    timeout_seconds: Optional[float] = None,
    max_retries: Optional[int] = None,
    lease_duration: float = 300,
    strict: bool = False,
    holder: "Optional[ConcurrencyLeaseHolder]" = None,
    suppress_warnings: bool = False,
) -> Generator[None, None, None]:
    """A context manager that acquires and releases concurrency slots from the
    given concurrency limits.

    Args:
        names: The names of the concurrency limits to acquire slots from.
        occupy: The number of slots to acquire and hold from each limit.
        timeout_seconds: The number of seconds to wait for the slots to be acquired before
            raising a `TimeoutError`. A timeout of `None` will wait indefinitely.
        max_retries: The maximum number of retries to acquire the concurrency slots.
        lease_duration: The duration of the lease for the acquired slots in seconds.
        strict: A boolean specifying whether to raise an error if the concurrency limit does not exist.
            Defaults to `False`.
        holder: A dictionary containing information about the holder of the concurrency slots.
            Typically includes 'type' and 'id' keys.

    Raises:
        TimeoutError: If the slots are not acquired within the given timeout.
        ConcurrencySlotAcquisitionError: If the concurrency limit does not exist and `strict` is `True`.

    Example:
    A simple example of using the sync `concurrency` context manager:
    ```python
    from prefect.concurrency.sync import concurrency

    def resource_heavy():
        with concurrency("test", occupy=1):
            print("Resource heavy task")

    def main():
        resource_heavy()
    ```
    """
    if not names:
        yield
        return

    names = names if isinstance(names, list) else [names]

    acquisition_response = acquire_concurrency_slots_with_lease(
        names,
        occupy,
        timeout_seconds=timeout_seconds,
        strict=strict,
        lease_duration=lease_duration,
        max_retries=max_retries,
        holder=holder,
        suppress_warnings=suppress_warnings,
    )

    if not acquisition_response.limits:
        yield
        return

    emitted_events = emit_concurrency_acquisition_events(
        acquisition_response.limits, occupy
    )

    try:
        with maintain_concurrency_lease(
            acquisition_response.lease_id,
            lease_duration,
            raise_on_lease_renewal_failure=strict,
            suppress_warnings=suppress_warnings,
        ):
            yield
    finally:
        release_concurrency_slots_with_lease(acquisition_response.lease_id)
        emit_concurrency_release_events(
            acquisition_response.limits, occupy, emitted_events
        )



"""
Utilities for extensions of and operations on Python collections.
"""

import io
import itertools
import types
from collections import OrderedDict
from collections.abc import (
    Callable,
    Collection,
    Generator,
    Hashable,
    Iterable,
    Iterator,
    Sequence,
    Set,
)
from dataclasses import fields, is_dataclass, replace
from enum import Enum, auto
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Optional,
    Union,
    cast,
    overload,
)
from unittest.mock import Mock

import pydantic
from typing_extensions import TypeAlias, TypeVar

# Quote moved to `prefect.utilities.annotations` but preserved here for compatibility
from prefect.utilities.annotations import BaseAnnotation as BaseAnnotation
from prefect.utilities.annotations import Quote as Quote
from prefect.utilities.annotations import opaque as opaque
from prefect.utilities.annotations import quote as quote

if TYPE_CHECKING:
    pass


class AutoEnum(str, Enum):
    """
    An enum class that automatically generates value from variable names.

    This guards against common errors where variable names are updated but values are
    not.

    In addition, because AutoEnums inherit from `str`, they are automatically
    JSON-serializable.

    See https://docs.python.org/3/library/enum.html#using-automatic-values

    Example:
        ```python
        class MyEnum(AutoEnum):
            RED = AutoEnum.auto() # equivalent to RED = 'RED'
            BLUE = AutoEnum.auto() # equivalent to BLUE = 'BLUE'
        ```
    """

    @staticmethod
    def _generate_next_value_(name: str, *_: object, **__: object) -> str:
        return name

    @staticmethod
    def auto() -> str:
        """
        Exposes `enum.auto()` to avoid requiring a second import to use `AutoEnum`
        """
        return auto()

    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.value}"


KT = TypeVar("KT")
VT = TypeVar("VT", infer_variance=True)
VT1 = TypeVar("VT1", infer_variance=True)
VT2 = TypeVar("VT2", infer_variance=True)
R = TypeVar("R", infer_variance=True)
NestedDict: TypeAlias = dict[KT, Union[VT, "NestedDict[KT, VT]"]]
HashableT = TypeVar("HashableT", bound=Hashable)


def dict_to_flatdict(dct: NestedDict[KT, VT]) -> dict[tuple[KT, ...], VT]:
    """Converts a (nested) dictionary to a flattened representation.

    Each key of the flat dict will be a CompoundKey tuple containing the "chain of keys"
    for the corresponding value.

    Args:
        dct (dict): The dictionary to flatten

    Returns:
        A flattened dict of the same type as dct
    """

    def flatten(
        dct: NestedDict[KT, VT], _parent: tuple[KT, ...] = ()
    ) -> Iterator[tuple[tuple[KT, ...], VT]]:
        parent = _parent or ()
        for k, v in dct.items():
            k_parent = (*parent, k)
            # if v is a non-empty dict, recurse
            if isinstance(v, dict) and v:
                yield from flatten(cast(NestedDict[KT, VT], v), _parent=k_parent)
            else:
                yield (k_parent, cast(VT, v))

    type_ = cast(type[dict[tuple[KT, ...], VT]], type(dct))
    return type_(flatten(dct))


def flatdict_to_dict(dct: dict[tuple[KT, ...], VT]) -> NestedDict[KT, VT]:
    """Converts a flattened dictionary back to a nested dictionary.

    Args:
        dct (dict): The dictionary to be nested. Each key should be a tuple of keys
            as generated by `dict_to_flatdict`

    Returns
        A nested dict of the same type as dct
    """

    type_ = cast(type[NestedDict[KT, VT]], type(dct))

    def new(type_: type[NestedDict[KT, VT]] = type_) -> NestedDict[KT, VT]:
        return type_()

    result = new()
    for key_tuple, value in dct.items():
        current = result
        *prefix_keys, last_key = key_tuple
        for prefix_key in prefix_keys:
            # Build nested dictionaries up for the current key tuple
            try:
                current = cast(NestedDict[KT, VT], current[prefix_key])
            except KeyError:
                new_dict = current[prefix_key] = new()
                current = new_dict

        # Set the value
        current[last_key] = value

    return result


T = TypeVar("T")


def isiterable(obj: Any) -> bool:
    """
    Return a boolean indicating if an object is iterable.

    Excludes types that are iterable but typically used as singletons:
    - str
    - bytes
    - IO objects
    """
    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return not isinstance(obj, (str, bytes, io.IOBase))


def ensure_iterable(obj: Union[T, Iterable[T]]) -> Collection[T]:
    if isinstance(obj, Sequence) or isinstance(obj, Set):
        return cast(Collection[T], obj)
    obj = cast(T, obj)  # No longer in the iterable case
    return [obj]


def listrepr(objs: Iterable[Any], sep: str = " ") -> str:
    return sep.join(repr(obj) for obj in objs)


def extract_instances(
    objects: Iterable[Any],
    types: Union[type[T], tuple[type[T], ...]] = object,
) -> Union[list[T], dict[type[T], list[T]]]:
    """
    Extract objects from a file and returns a dict of type -> instances

    Args:
        objects: An iterable of objects
        types: A type or tuple of types to extract, defaults to all objects

    Returns:
        If a single type is given: a list of instances of that type
        If a tuple of types is given: a mapping of type to a list of instances
    """
    types_collection = ensure_iterable(types)

    # Create a mapping of type -> instance from the exec values
    ret: dict[type[T], list[Any]] = {}

    for o in objects:
        # We iterate here so that the key is the passed type rather than type(o)
        for type_ in types_collection:
            if isinstance(o, type_):
                ret.setdefault(type_, []).append(o)

    if len(types_collection) == 1:
        [type_] = types_collection
        return ret[type_]

    return ret


def batched_iterable(
    iterable: Iterable[T], size: int
) -> Generator[tuple[T, ...], None, None]:
    """
    Yield batches of a certain size from an iterable

    Args:
        iterable (Iterable): An iterable
        size (int): The batch size to return

    Yields:
        tuple: A batch of the iterable
    """
    it = iter(iterable)
    while True:
        batch = tuple(itertools.islice(it, size))
        if not batch:
            break
        yield batch


class StopVisiting(BaseException):
    """
    A special exception used to stop recursive visits in `visit_collection`.

    When raised, the expression is returned without modification and recursive visits
    in that path will end.
    """


@overload
def visit_collection(
    expr: Any,
    visit_fn: Callable[[Any, dict[str, VT]], Any],
    *,
    return_data: Literal[True] = ...,
    max_depth: int = ...,
    context: dict[str, VT] = ...,
    remove_annotations: bool = ...,
    _seen: Optional[dict[int, Any]] = ...,
) -> Any: ...


@overload
def visit_collection(
    expr: Any,
    visit_fn: Callable[[Any], Any],
    *,
    return_data: Literal[True] = ...,
    max_depth: int = ...,
    context: None = None,
    remove_annotations: bool = ...,
    _seen: Optional[dict[int, Any]] = ...,
) -> Any: ...


@overload
def visit_collection(
    expr: Any,
    visit_fn: Callable[[Any, dict[str, VT]], Any],
    *,
    return_data: bool = ...,
    max_depth: int = ...,
    context: dict[str, VT] = ...,
    remove_annotations: bool = ...,
    _seen: Optional[dict[int, Any]] = ...,
) -> Optional[Any]: ...


@overload
def visit_collection(
    expr: Any,
    visit_fn: Callable[[Any], Any],
    *,
    return_data: bool = ...,
    max_depth: int = ...,
    context: None = None,
    remove_annotations: bool = ...,
    _seen: Optional[dict[int, Any]] = ...,
) -> Optional[Any]: ...


@overload
def visit_collection(
    expr: Any,
    visit_fn: Callable[[Any, dict[str, VT]], Any],
    *,
    return_data: Literal[False] = False,
    max_depth: int = ...,
    context: dict[str, VT] = ...,
    remove_annotations: bool = ...,
    _seen: Optional[dict[int, Any]] = ...,
) -> None: ...


def visit_collection(
    expr: Any,
    visit_fn: Union[Callable[[Any, dict[str, VT]], Any], Callable[[Any], Any]],
    *,
    return_data: bool = False,
    max_depth: int = -1,
    context: Optional[dict[str, VT]] = None,
    remove_annotations: bool = False,
    _seen: Optional[dict[int, Any]] = None,
) -> Optional[Any]:
    """
    Visits and potentially transforms every element of an arbitrary Python collection.

    If an element is a Python collection, it will be visited recursively. If an element
    is not a collection, `visit_fn` will be called with the element. The return value of
    `visit_fn` can be used to alter the element if `return_data` is set to `True`.

    Note:
    - When `return_data` is `True`, a copy of each collection is created only if
      `visit_fn` modifies an element within that collection. This approach minimizes
      performance penalties by avoiding unnecessary copying.
    - When `return_data` is `False`, no copies are created, and only side effects from
      `visit_fn` are applied. This mode is faster and should be used when no transformation
      of the collection is required, because it never has to copy any data.

    Supported types:
    - List (including iterators)
    - Tuple
    - Set
    - Dict (note: keys are also visited recursively)
    - Dataclass
    - Pydantic model
    - Prefect annotations

    Note that visit_collection will not consume generators or async generators, as it would prevent
    the caller from iterating over them.

    Args:
        expr (Any): A Python object or expression.
        visit_fn (Callable[[Any, Optional[dict]], Any] or Callable[[Any], Any]): A function
            that will be applied to every non-collection element of `expr`. The function can
            accept one or two arguments. If two arguments are accepted, the second argument
            will be the context dictionary.
        return_data (bool): If `True`, a copy of `expr` containing data modified by `visit_fn`
            will be returned. This is slower than `return_data=False` (the default).
        max_depth (int): Controls the depth of recursive visitation. If set to zero, no
            recursion will occur. If set to a positive integer `N`, visitation will only
            descend to `N` layers deep. If set to any negative integer, no limit will be
            enforced and recursion will continue until terminal items are reached. By
            default, recursion is unlimited.
        context (Optional[dict]): An optional dictionary. If passed, the context will be sent
            to each call to the `visit_fn`. The context can be mutated by each visitor and
            will be available for later visits to expressions at the given depth. Values
            will not be available "up" a level from a given expression.
            The context will be automatically populated with an 'annotation' key when
            visiting collections within a `BaseAnnotation` type. This requires the caller to
            pass `context={}` and will not be activated by default.
        remove_annotations (bool): If set, annotations will be replaced by their contents. By
            default, annotations are preserved but their contents are visited.
        _seen (Optional[Set[int]]): A set of object ids that have already been visited. This
            prevents infinite recursion when visiting recursive data structures.

    Returns:
        Any: The modified collection if `return_data` is `True`, otherwise `None`.
    """

    if _seen is None:
        _seen = {}

    if context is not None:
        _callback = cast(Callable[[Any, dict[str, VT]], Any], visit_fn)

        def visit_nested(expr: Any) -> Optional[Any]:
            return visit_collection(
                expr,
                _callback,
                return_data=return_data,
                remove_annotations=remove_annotations,
                max_depth=max_depth - 1,
                # Copy the context on nested calls so it does not "propagate up"
                context=context.copy(),
                _seen=_seen,
            )

        def visit_expression(expr: Any) -> Any:
            return _callback(expr, context)
    else:
        _callback = cast(Callable[[Any], Any], visit_fn)

        def visit_nested(expr: Any) -> Optional[Any]:
            # Utility for a recursive call, preserving options and updating the depth.
            return visit_collection(
                expr,
                _callback,
                return_data=return_data,
                remove_annotations=remove_annotations,
                max_depth=max_depth - 1,
                _seen=_seen,
            )

        def visit_expression(expr: Any) -> Any:
            return _callback(expr)

    # --- 1. Visit every expression
    try:
        result = visit_expression(expr)
    except StopVisiting:
        max_depth = 0
        result = expr

    if return_data:
        # Only mutate the root expression if the user indicated we're returning data,
        # otherwise the function could return null and we have no collection to check
        expr = result

    # --- 2. Visit every child of the expression recursively

    # If we have reached the maximum depth or we have already visited this object,
    # return the result if we are returning data, otherwise return None
    obj_id = id(expr)
    if max_depth == 0:
        return result if return_data else None
    elif obj_id in _seen:
        # Return the cached transformed result
        return _seen[obj_id] if return_data else None

    # Mark this object as being processed to handle circular references
    # We'll update with the actual result later
    _seen[obj_id] = expr

    # Then visit every item in the expression if it is a collection

    # presume that the result is the original expression.
    # in each of the following cases, we will update the result if we need to.
    result = expr

    # --- Generators

    if isinstance(expr, (types.GeneratorType, types.AsyncGeneratorType)):
        # Do not attempt to iterate over generators, as it will exhaust them
        pass

    # --- Mocks

    elif isinstance(expr, Mock):
        # Do not attempt to recurse into mock objects
        pass

    # --- Annotations (unmapped, quote, etc.)

    elif isinstance(expr, BaseAnnotation):
        annotated = cast(BaseAnnotation[Any], expr)
        if context is not None:
            context["annotation"] = cast(VT, annotated)
        unwrapped = annotated.unwrap()

        if isinstance(annotated, opaque):
            try:
                value = visit_expression(unwrapped)
            except StopVisiting:
                value = unwrapped
        else:
            value = visit_nested(unwrapped)

        if return_data:
            # if we are removing annotations, return the value
            if remove_annotations:
                result = value
            # if the value was modified, rewrap it
            elif value is not unwrapped:
                result = annotated.rewrap(value)
            # otherwise return the expr

    # --- Sequences

    elif isinstance(expr, (list, tuple, set)):
        seq = cast(Union[list[Any], tuple[Any], set[Any]], expr)
        items = [visit_nested(o) for o in seq]
        if return_data:
            modified = any(item is not orig for item, orig in zip(items, seq))
            if modified:
                # Use _make for NamedTuples
                if hasattr(type(seq), "_make"):
                    result = type(seq)._make(items)
                else:
                    result = type(seq)(items)

    # --- Dictionaries

    elif isinstance(expr, (dict, OrderedDict)):
        mapping = cast(dict[Any, Any], expr)
        items = [(visit_nested(k), visit_nested(v)) for k, v in mapping.items()]
        if return_data:
            modified = any(
                k1 is not k2 or v1 is not v2
                for (k1, v1), (k2, v2) in zip(items, mapping.items())
            )
            if modified:
                result = type(mapping)(items)

    # --- Dataclasses

    elif is_dataclass(expr) and not isinstance(expr, type):
        expr_fields = fields(expr)
        values = [visit_nested(getattr(expr, f.name)) for f in expr_fields]
        if return_data:
            modified = any(
                getattr(expr, f.name) is not v for f, v in zip(expr_fields, values)
            )
            if modified:
                result = replace(
                    expr, **{f.name: v for f, v in zip(expr_fields, values)}
                )

    # --- Pydantic models

    elif isinstance(expr, pydantic.BaseModel):
        # when extra=allow, fields not in model_fields may be in model_fields_set
        original_data = dict(expr)
        updated_data = {
            field: visit_nested(value) for field, value in original_data.items()
        }

        if return_data:
            modified = any(
                original_data[field] is not updated_data[field]
                for field in updated_data
            )
            if modified:
                # Use construct to avoid validation and handle immutability
                model_instance = expr.model_construct(
                    _fields_set=expr.model_fields_set, **updated_data
                )
                for private_attr in expr.__private_attributes__:
                    setattr(model_instance, private_attr, getattr(expr, private_attr))
                result = model_instance

    # Update the cache with the final transformed result
    if return_data:
        _seen[obj_id] = result

    if return_data:
        return result


@overload
def remove_nested_keys(
    keys_to_remove: list[HashableT], obj: NestedDict[HashableT, VT]
) -> NestedDict[HashableT, VT]: ...


@overload
def remove_nested_keys(keys_to_remove: list[HashableT], obj: Any) -> Any: ...


def remove_nested_keys(
    keys_to_remove: list[HashableT], obj: Union[NestedDict[HashableT, VT], Any]
) -> Union[NestedDict[HashableT, VT], Any]:
    """
    Recurses a dictionary returns a copy without all keys that match an entry in
    `key_to_remove`. Return `obj` unchanged if not a dictionary.

    Args:
        keys_to_remove: A list of keys to remove from obj obj: The object to remove keys
            from.

    Returns:
        `obj` without keys matching an entry in `keys_to_remove` if `obj` is a
            dictionary. `obj` if `obj` is not a dictionary.
    """
    if not isinstance(obj, dict):
        return obj
    return {
        key: remove_nested_keys(keys_to_remove, value)
        for key, value in cast(NestedDict[HashableT, VT], obj).items()
        if key not in keys_to_remove
    }


@overload
def distinct(
    iterable: Iterable[HashableT], key: None = None
) -> Iterator[HashableT]: ...


@overload
def distinct(iterable: Iterable[T], key: Callable[[T], Hashable]) -> Iterator[T]: ...


def distinct(
    iterable: Iterable[Union[T, HashableT]],
    key: Optional[Callable[[T], Hashable]] = None,
) -> Iterator[Union[T, HashableT]]:
    def _key(__i: Any) -> Hashable:
        return __i

    if key is not None:
        _key = cast(Callable[[Any], Hashable], key)

    seen: set[Hashable] = set()
    for item in iterable:
        if _key(item) in seen:
            continue
        seen.add(_key(item))
        yield item


@overload
def get_from_dict(
    dct: NestedDict[str, VT], keys: Union[str, list[str]], default: None = None
) -> Optional[VT]: ...


@overload
def get_from_dict(
    dct: NestedDict[str, VT], keys: Union[str, list[str]], default: R
) -> Union[VT, R]: ...


def get_from_dict(
    dct: NestedDict[str, VT], keys: Union[str, list[str]], default: Optional[R] = None
) -> Union[VT, R, None]:
    """
    Fetch a value from a nested dictionary or list using a sequence of keys.

    This function allows to fetch a value from a deeply nested structure
    of dictionaries and lists using either a dot-separated string or a list
    of keys. If a requested key does not exist, the function returns the
    provided default value.

    Args:
        dct: The nested dictionary or list from which to fetch the value.
        keys: The sequence of keys to use for access. Can be a
            dot-separated string or a list of keys. List indices can be included
            in the sequence as either integer keys or as string indices in square
            brackets.
        default: The default value to return if the requested key path does not
            exist. Defaults to None.

    Returns:
        The fetched value if the key exists, or the default value if it does not.

    Examples:

    ```python
    get_from_dict({'a': {'b': {'c': [1, 2, 3, 4]}}}, 'a.b.c[1]') # 2
    get_from_dict({'a': {'b': [0, {'c': [1, 2]}]}}, ['a', 'b', 1, 'c', 1]) # 2
    get_from_dict({'a': {'b': [0, {'c': [1, 2]}]}}, 'a.b.1.c.2', 'default') # 'default'
    ```
    """
    if isinstance(keys, str):
        keys = keys.replace("[", ".").replace("]", "").split(".")
    value = dct
    try:
        for key in keys:
            try:
                # Try to cast to int to handle list indices
                key = int(key)
            except ValueError:
                # If it's not an int, use the key as-is
                # for dict lookup
                pass
            value = value[key]  # type: ignore
        return cast(VT, value)
    except (TypeError, KeyError, IndexError):
        return default


def set_in_dict(
    dct: NestedDict[str, VT], keys: Union[str, list[str]], value: VT
) -> None:
    """
    Sets a value in a nested dictionary using a sequence of keys.

    This function allows to set a value in a deeply nested structure
    of dictionaries and lists using either a dot-separated string or a list
    of keys. If a requested key does not exist, the function will create it as
    a new dictionary.

    Args:
        dct: The dictionary to set the value in.
        keys: The sequence of keys to use for access. Can be a
            dot-separated string or a list of keys.
        value: The value to set in the dictionary.

    Returns:
        The modified dictionary with the value set at the specified key path.

    Raises:
        KeyError: If the key path exists and is not a dictionary.
    """
    if isinstance(keys, str):
        keys = keys.replace("[", ".").replace("]", "").split(".")
    for k in keys[:-1]:
        if not isinstance(dct.get(k, {}), dict):
            raise TypeError(f"Key path exists and contains a non-dict value: {keys}")
        if k not in dct:
            dct[k] = {}
        dct = cast(NestedDict[str, VT], dct[k])
    dct[keys[-1]] = value


def deep_merge(
    dct: NestedDict[str, VT1], merge: NestedDict[str, VT2]
) -> NestedDict[str, Union[VT1, VT2]]:
    """
    Recursively merges `merge` into `dct`.

    Args:
        dct: The dictionary to merge into.
        merge: The dictionary to merge from.

    Returns:
        A new dictionary with the merged contents.
    """
    result: dict[str, Any] = dct.copy()  # Start with keys and values from `dct`
    for key, value in merge.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            # If both values are dictionaries, merge them recursively
            result[key] = deep_merge(
                cast(NestedDict[str, VT1], result[key]),
                cast(NestedDict[str, VT2], value),
            )
        else:
            # Otherwise, overwrite with the new value
            result[key] = cast(Union[VT2, NestedDict[str, VT2]], value)
    return result


def deep_merge_dicts(*dicts: NestedDict[str, Any]) -> NestedDict[str, Any]:
    """
    Recursively merges multiple dictionaries.

    Args:
        dicts: The dictionaries to merge.

    Returns:
        A new dictionary with the merged contents.
    """
    result: NestedDict[str, Any] = {}
    for dictionary in dicts:
        result = deep_merge(result, dictionary)
    return result



import warnings
from typing import ClassVar, Optional

from pydantic import Field
from pydantic_settings import SettingsConfigDict

from prefect.settings.base import PrefectBaseSettings, build_settings_config
from prefect.types import LogLevel


class RunnerServerSettings(PrefectBaseSettings):
    """
    Settings for controlling runner server behavior
    """

    model_config: ClassVar[SettingsConfigDict] = build_settings_config(
        ("runner", "server")
    )

    enable: bool = Field(
        default=False,
        description="Whether or not to enable the runner's webserver.",
    )

    host: str = Field(
        default="localhost",
        description="The host address the runner's webserver should bind to.",
    )

    port: int = Field(
        default=8080,
        description="The port the runner's webserver should bind to.",
    )

    log_level: LogLevel = Field(
        default="ERROR",
        description="The log level of the runner's webserver.",
    )

    missed_polls_tolerance: int = Field(
        default=2,
        description="Number of missed polls before a runner is considered unhealthy by its webserver.",
    )


class RunnerSettings(PrefectBaseSettings):
    """
    Settings for controlling runner behavior
    """

    model_config: ClassVar[SettingsConfigDict] = build_settings_config(("runner",))

    process_limit: int = Field(
        default=5,
        description="Maximum number of processes a runner will execute in parallel.",
    )

    poll_frequency: int = Field(
        default=10,
        description="Number of seconds a runner should wait between queries for scheduled work.",
    )

    crash_on_cancellation_failure: bool = Field(
        default=False,
        description=(
            "Whether to crash flow runs and shut down the runner when cancellation "
            "observing fails. When enabled, if both websocket and polling mechanisms "
            "for detecting cancellation events fail, all in-flight flow runs will be "
            "marked as crashed and the runner will shut down. When disabled (default), "
            "the runner will log an error but continue executing flow runs."
        ),
    )

    server: RunnerServerSettings = Field(
        default_factory=RunnerServerSettings,
        description="Settings for controlling runner server behavior",
    )

    # handle deprecated fields

    @property
    def heartbeat_frequency(self) -> Optional[int]:
        """Deprecated: Use flows.heartbeat_frequency instead."""
        from prefect.settings.context import get_current_settings

        warnings.warn(
            "`runner.heartbeat_frequency` has been moved to `flows.heartbeat_frequency`. "
            "Use `PREFECT_FLOWS_HEARTBEAT_FREQUENCY` instead of `PREFECT_RUNNER_HEARTBEAT_FREQUENCY`.",
            DeprecationWarning,
            stacklevel=2,
        )
        return get_current_settings().flows.heartbeat_frequency



from __future__ import annotations

import asyncio
import logging
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, AsyncGenerator, Literal, Optional
from uuid import UUID

import anyio
import httpx

from prefect.client.orchestration import get_client
from prefect.client.schemas.responses import (
    ConcurrencyLimitWithLeaseResponse,
    MinimalConcurrencyLimitResponse,
)
from prefect.concurrency._events import (
    emit_concurrency_acquisition_events,
    emit_concurrency_release_events,
)
from prefect.concurrency._leases import amaintain_concurrency_lease
from prefect.concurrency.context import ConcurrencyContext
from prefect.logging import get_logger
from prefect.logging.loggers import get_run_logger

from .services import (
    ConcurrencySlotAcquisitionService,
    ConcurrencySlotAcquisitionWithLeaseService,
)

if TYPE_CHECKING:
    from prefect.client.schemas.objects import ConcurrencyLeaseHolder


class ConcurrencySlotAcquisitionError(Exception):
    """Raised when an unhandlable occurs while acquiring concurrency slots."""


class AcquireConcurrencySlotTimeoutError(TimeoutError):
    """Raised when acquiring a concurrency slot times out."""


logger: logging.Logger = get_logger("concurrency")


async def aacquire_concurrency_slots(
    names: list[str],
    slots: int,
    mode: Literal["concurrency", "rate_limit"] = "concurrency",
    timeout_seconds: Optional[float] = None,
    max_retries: Optional[int] = None,
    strict: bool = False,
) -> list[MinimalConcurrencyLimitResponse]:
    service = ConcurrencySlotAcquisitionService.instance(frozenset(names))
    future = service.send((slots, mode, timeout_seconds, max_retries))
    try:
        response = await asyncio.wrap_future(future)
    except TimeoutError as timeout:
        raise AcquireConcurrencySlotTimeoutError(
            f"Attempt to acquire concurrency slots timed out after {timeout_seconds} second(s)"
        ) from timeout
    except Exception as exc:
        raise ConcurrencySlotAcquisitionError(
            f"Unable to acquire concurrency slots on {names!r}"
        ) from exc

    retval = _response_to_minimal_concurrency_limit_response(response)

    if not retval:
        if strict:
            raise ConcurrencySlotAcquisitionError(
                f"Concurrency limits {names!r} must be created before acquiring slots"
            )
        try:
            logger = get_run_logger()
        except Exception:
            pass
        else:
            logger.warning(
                f"Concurrency limits {names!r} do not exist - skipping acquisition."
            )

    return retval


async def aacquire_concurrency_slots_with_lease(
    names: list[str],
    slots: int,
    mode: Literal["concurrency", "rate_limit"] = "concurrency",
    timeout_seconds: Optional[float] = None,
    max_retries: Optional[int] = None,
    lease_duration: float = 300,
    strict: bool = False,
    holder: "Optional[ConcurrencyLeaseHolder]" = None,
    suppress_warnings: bool = False,
) -> ConcurrencyLimitWithLeaseResponse:
    service = ConcurrencySlotAcquisitionWithLeaseService.instance(frozenset(names))
    future = service.send(
        (slots, mode, timeout_seconds, max_retries, lease_duration, strict, holder)
    )
    try:
        response = await asyncio.wrap_future(future)
    except TimeoutError as timeout:
        raise AcquireConcurrencySlotTimeoutError(
            f"Attempt to acquire concurrency slots timed out after {timeout_seconds} second(s)"
        ) from timeout
    except Exception as exc:
        raise ConcurrencySlotAcquisitionError(
            f"Unable to acquire concurrency slots on {names!r}"
        ) from exc

    retval = ConcurrencyLimitWithLeaseResponse.model_validate(response.json())

    if not retval.limits:
        if strict:
            raise ConcurrencySlotAcquisitionError(
                f"Concurrency limits {names!r} must be created before acquiring slots"
            )
        else:
            try:
                # Use a run logger if available
                task_logger = get_run_logger()
            except Exception:
                task_logger = get_logger("concurrency")

            log_level = logging.DEBUG if suppress_warnings else logging.WARNING
            task_logger.log(
                log_level,
                f"Concurrency limits {names!r} do not exist - skipping acquisition.",
            )

    return retval


async def arelease_concurrency_slots(
    names: list[str], slots: int, occupancy_seconds: float
) -> list[MinimalConcurrencyLimitResponse]:
    async with get_client() as client:
        response = await client.release_concurrency_slots(
            names=names, slots=slots, occupancy_seconds=occupancy_seconds
        )
        return _response_to_minimal_concurrency_limit_response(response)


async def arelease_concurrency_slots_with_lease(
    lease_id: UUID,
) -> None:
    async with get_client() as client:
        await client.release_concurrency_slots_with_lease(lease_id=lease_id)


def _response_to_minimal_concurrency_limit_response(
    response: httpx.Response,
) -> list[MinimalConcurrencyLimitResponse]:
    return [
        MinimalConcurrencyLimitResponse.model_validate(obj_) for obj_ in response.json()
    ]


@asynccontextmanager
async def concurrency(
    names: str | list[str],
    occupy: int = 1,
    timeout_seconds: Optional[float] = None,
    max_retries: Optional[int] = None,
    lease_duration: float = 300,
    strict: bool = False,
    holder: "Optional[ConcurrencyLeaseHolder]" = None,
    suppress_warnings: bool = False,
) -> AsyncGenerator[None, None]:
    """
    Internal version of the `concurrency` context manager. The public version is located in `prefect.concurrency.asyncio`.

    Args:
        names: The names of the concurrency limits to acquire slots from.
        occupy: The number of slots to acquire and hold from each limit.
        timeout_seconds: The number of seconds to wait for the slots to be acquired before
            raising a `TimeoutError`. A timeout of `None` will wait indefinitely.
        max_retries: The maximum number of retries to acquire the concurrency slots.
        lease_duration: The duration of the lease for the acquired slots in seconds.
        strict: A boolean specifying whether to raise an error if the concurrency limit does not exist.
            Defaults to `False`.
        holder: A dictionary containing information about the holder of the concurrency slots.
            Typically includes 'type' and 'id' keys.

    Raises:
        TimeoutError: If the slots are not acquired within the given timeout.
        ConcurrencySlotAcquisitionError: If the concurrency limit does not exist and `strict` is `True`.

    Example:
    A simple example of using the async `concurrency` context manager:
    ```python
    from prefect.concurrency.asyncio import concurrency

    async def resource_heavy():
        async with concurrency("test", occupy=1):
            print("Resource heavy task")

    async def main():
        await resource_heavy()
    ```
    """
    if not names:
        yield
        return

    names = names if isinstance(names, list) else [names]

    response = await aacquire_concurrency_slots_with_lease(
        names=names,
        slots=occupy,
        timeout_seconds=timeout_seconds,
        max_retries=max_retries,
        lease_duration=lease_duration,
        strict=strict,
        holder=holder,
        suppress_warnings=suppress_warnings,
    )

    if not response.limits:
        yield
        return

    emitted_events = emit_concurrency_acquisition_events(response.limits, occupy)

    try:
        async with amaintain_concurrency_lease(
            response.lease_id,
            lease_duration,
            raise_on_lease_renewal_failure=strict,
            suppress_warnings=suppress_warnings,
        ):
            yield
    finally:
        try:
            await arelease_concurrency_slots_with_lease(
                lease_id=response.lease_id,
            )
        except anyio.get_cancelled_exc_class():
            # The task was cancelled before it could release the lease. Add the
            # lease ID to the cleanup list so it can be released when the
            # concurrency context is exited.
            if ctx := ConcurrencyContext.get():
                ctx.cleanup_lease_ids.append(response.lease_id)

        emit_concurrency_release_events(response.limits, occupy, emitted_events)



from __future__ import annotations

from datetime import timedelta
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from prefect._experimental.sla.objects import SlaTypes
from prefect.client.schemas.actions import DeploymentScheduleCreate
from prefect.client.schemas.schedules import SCHEDULE_TYPES


class WorkPoolConfig(BaseModel):
    model_config = ConfigDict(extra="ignore")

    name: Optional[str] = None
    work_queue_name: Optional[str] = None
    job_variables: Dict[str, Any] = Field(default_factory=dict)


class DeploymentConfig(BaseModel):
    model_config = ConfigDict(extra="ignore")

    # base metadata
    name: Optional[str] = None
    version: Optional[str] = None
    version_type: Optional[str] = None
    tags: Optional[Union[str, list[Any]]] = (
        None  # allow raw templated string or list; templating will normalize
    )
    description: Optional[str] = None

    # schedule metadata
    schedule: Optional["ScheduleItem"] = None
    schedules: Optional[List["ScheduleItem"]] = None
    paused: Optional[bool] = None
    concurrency_limit: Optional[Union[int, "ConcurrencyLimitSpec"]] = None

    # flow-specific
    flow_name: Optional[str] = None
    entrypoint: Optional[str] = None
    parameters: Dict[str, Any] = Field(default_factory=dict)
    enforce_parameter_schema: Optional[bool] = None

    # per-deployment actions (optional overrides)
    # Accept list, mapping (empty or step), or null for flexibility
    build: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]] = None
    push: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]] = None
    pull: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]] = None

    # infra-specific
    work_pool: Optional[WorkPoolConfig] = None

    # automations metadata
    # Triggers are stored as raw dicts to allow Jinja templating (e.g., enabled: "{{ prefect.variables.is_prod }}")
    # Strict validation happens later in _initialize_deployment_triggers after template resolution
    triggers: Optional[List[Dict[str, Any]]] = None
    sla: Optional[List[SlaTypes]] = None


class PrefectYamlModel(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="ignore")

    # generic metadata (currently unused by CLI but allowed)
    prefect_version: Optional[str] = Field(default=None, alias="prefect-version")
    name: Optional[str] = None

    # global actions
    build: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]] = None
    push: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]] = None
    pull: Optional[Union[List[Dict[str, Any]], Dict[str, Any]]] = None

    # deployments
    deployments: List[DeploymentConfig] = Field(default_factory=list)

    @staticmethod
    def _validate_action_steps(steps: Optional[List[Dict[str, Any]]]) -> None:
        # Light validation: allow any mapping; prefer single-key style but do not enforce
        if not steps:
            return
        for step in steps:
            if not isinstance(step, dict):
                raise TypeError("Each action step must be a mapping")
            # empty or multi-key steps will be passed through unchanged

    @field_validator("build", "push", "pull")
    @classmethod
    def _validate_actions(cls, v: Optional[List[Dict[str, Any]]]):
        cls._validate_action_steps(v)
        return v

    @field_validator("deployments")
    @classmethod
    def _validate_deployments(cls, v: List[DeploymentConfig]):
        # Ensure deployments is a list
        return v or []


class ConcurrencyLimitSpec(BaseModel):
    model_config = ConfigDict(extra="ignore")

    limit: Optional[int] = None
    collision_strategy: Optional[str] = None
    grace_period_seconds: Optional[int] = None


class RawScheduleConfig(BaseModel):
    """
    Strongly-typed schedule config that mirrors the CLI's accepted YAML shape.
    Exactly one of cron, interval, or rrule must be provided.
    """

    model_config = ConfigDict(extra="forbid")

    # One-of schedule selectors
    cron: Optional[str] = None
    interval: Optional[timedelta] = (
        None  # accepts int/float (seconds), ISO 8601, HH:MM:SS
    )
    rrule: Optional[str] = None

    # Common extras
    timezone: Optional[str] = None
    anchor_date: Optional[str] = None
    active: Optional[Union[bool, str]] = None  # Allow string for template values
    parameters: Dict[str, Any] = Field(default_factory=dict)
    slug: Optional[str] = None
    replaces: Optional[str] = None  # The slug of an existing schedule to replace

    # Cron-specific
    day_or: Optional[Union[bool, str]] = None  # Allow string for template values

    @model_validator(mode="after")
    def _one_of_schedule(self):
        provided = [v is not None for v in (self.cron, self.interval, self.rrule)]
        if sum(provided) != 1:
            raise ValueError(
                "Exactly one of 'cron', 'interval', or 'rrule' must be provided"
            )
        return self


ScheduleItem = Union[
    RawScheduleConfig, DeploymentScheduleCreate, SCHEDULE_TYPES, Dict[None, None]
]



import os
import sys
import warnings
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Type, get_origin

import dotenv
from cachetools import TTLCache
from pydantic import AliasChoices
from pydantic.fields import FieldInfo
from pydantic_settings import (
    BaseSettings,
    DotEnvSettingsSource,
    EnvSettingsSource,
    PydanticBaseSettingsSource,
)
from pydantic_settings.sources import (
    ENV_FILE_SENTINEL,
    ConfigFileSourceMixin,
    DotenvType,
)

from prefect._internal.compatibility.backports import tomllib
from prefect.settings.constants import DEFAULT_PREFECT_HOME, DEFAULT_PROFILES_PATH
from prefect.utilities.collections import get_from_dict

_file_cache: TTLCache[str, dict[str, Any]] = TTLCache(maxsize=100, ttl=60)


def _read_toml_file(path: Path) -> dict[str, Any]:
    """use ttl cache to cache toml files"""
    modified_time = path.stat().st_mtime
    cache_key = f"toml_file:{path}:{modified_time}"
    if value := _file_cache.get(cache_key):
        return value
    data = tomllib.loads(path.read_text(encoding="utf-8"))  # type: ignore
    _file_cache[cache_key] = data
    return data


class EnvFilterSettingsSource(EnvSettingsSource):
    """
    Custom pydantic settings source to filter out specific environment variables.

    All validation aliases are loaded from environment variables by default. We use
    `AliasPath` to maintain the ability set fields via model initialization, but those
    shouldn't be loaded from environment variables. This loader allows use to say which
    environment variables should be ignored.
    """

    def __init__(
        self,
        settings_cls: type[BaseSettings],
        case_sensitive: Optional[bool] = None,
        env_prefix: Optional[str] = None,
        env_nested_delimiter: Optional[str] = None,
        env_ignore_empty: Optional[bool] = None,
        env_parse_none_str: Optional[str] = None,
        env_parse_enums: Optional[bool] = None,
        env_filter: Optional[List[str]] = None,
    ) -> None:
        super().__init__(
            settings_cls,
            case_sensitive,
            env_prefix,
            env_nested_delimiter,
            env_ignore_empty,
            env_parse_none_str,
            env_parse_enums,
        )
        self.env_vars: Mapping[str, str | None]
        if env_filter:
            if isinstance(self.env_vars, dict):
                for key in env_filter:
                    self.env_vars.pop(key, None)
            else:
                self.env_vars = {
                    key: value
                    for key, value in self.env_vars.items()  # type: ignore
                    if key.lower() not in env_filter
                }


class FilteredDotEnvSettingsSource(DotEnvSettingsSource):
    def __init__(
        self,
        settings_cls: type[BaseSettings],
        env_file: Optional[DotenvType] = ENV_FILE_SENTINEL,
        env_file_encoding: Optional[str] = None,
        case_sensitive: Optional[bool] = None,
        env_prefix: Optional[str] = None,
        env_nested_delimiter: Optional[str] = None,
        env_ignore_empty: Optional[bool] = None,
        env_parse_none_str: Optional[str] = None,
        env_parse_enums: Optional[bool] = None,
        env_blacklist: Optional[List[str]] = None,
    ) -> None:
        super().__init__(
            settings_cls,
            env_file,
            env_file_encoding,
            case_sensitive,
            env_prefix,
            env_nested_delimiter,
            env_ignore_empty,
            env_parse_none_str,
            env_parse_enums,
        )
        self.env_blacklist = env_blacklist
        if self.env_blacklist:
            if isinstance(self.env_vars, dict):
                for key in self.env_blacklist:
                    self.env_vars.pop(key, None)
            else:
                self.env_vars: dict[str, str | None] = {
                    key: value
                    for key, value in self.env_vars.items()  # type: ignore
                    if key.lower() not in env_blacklist
                }


class ProfileSettingsTomlLoader(PydanticBaseSettingsSource):
    """
    Custom pydantic settings source to load profile settings from a toml file.

    See https://docs.pydantic.dev/latest/concepts/pydantic_settings/#customise-settings-sources
    """

    def __init__(self, settings_cls: Type[BaseSettings]):
        super().__init__(settings_cls)
        self.settings_cls = settings_cls
        self.profiles_path: Path = _get_profiles_path()
        self.profile_settings: dict[str, Any] = self._load_profile_settings()

    def _load_profile_settings(self) -> Dict[str, Any]:
        """Helper method to load the profile settings from the profiles.toml file"""
        if not self.profiles_path.exists():
            return self._get_default_profile()

        try:
            all_profile_data = _read_toml_file(self.profiles_path)
        except tomllib.TOMLDecodeError:
            warnings.warn(
                f"Failed to load profiles from {self.profiles_path}. Please ensure the file is valid TOML."
            )
            return {}

        if (
            sys.argv[0].endswith("/prefect")
            and len(sys.argv) >= 3
            and sys.argv[1] == "--profile"
        ):
            active_profile = sys.argv[2]

        else:
            active_profile = os.environ.get("PREFECT_PROFILE") or all_profile_data.get(
                "active"
            )

        profiles_data = all_profile_data.get("profiles", {})

        if not active_profile or active_profile not in profiles_data:
            return self._get_default_profile()
        return profiles_data[active_profile]

    def _get_default_profile(self) -> Dict[str, Any]:
        """Helper method to get the default profile"""
        default_profile_data = _read_toml_file(DEFAULT_PROFILES_PATH)
        default_profile = default_profile_data.get("active", "ephemeral")
        assert isinstance(default_profile, str)
        return default_profile_data.get("profiles", {}).get(default_profile, {})

    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> Tuple[Any, str, bool]:
        """Concrete implementation to get the field value from the profile settings"""
        if field.validation_alias:
            # Use validation alias as the key to ensure profile value does not
            # higher priority sources. Lower priority sources that use the
            # field name can override higher priority sources that use the
            # validation alias as seen in https://github.com/PrefectHQ/prefect/issues/15981
            if isinstance(field.validation_alias, str):
                value = self.profile_settings.get(field.validation_alias.upper())
                if value is not None:
                    return value, field.validation_alias, self.field_is_complex(field)
            elif isinstance(field.validation_alias, AliasChoices):
                value = None
                lowest_priority_alias = next(
                    choice
                    for choice in reversed(field.validation_alias.choices)
                    if isinstance(choice, str)
                )
                for alias in field.validation_alias.choices:
                    if not isinstance(alias, str):
                        continue
                    value = self.profile_settings.get(alias.upper())
                    if value is not None:
                        return (
                            value,
                            lowest_priority_alias,
                            self.field_is_complex(field),
                        )

        name = f"{self.config.get('env_prefix', '')}{field_name.upper()}"
        value = self.profile_settings.get(name)
        return value, field_name, self.field_is_complex(field)

    def __call__(self) -> Dict[str, Any]:
        """Called by pydantic to get the settings from our custom source"""
        if _is_test_mode():
            return {}
        profile_settings: Dict[str, Any] = {}
        for field_name, field in self.settings_cls.model_fields.items():
            value, key, is_complex = self.get_field_value(field, field_name)
            if value is not None:
                prepared_value = self.prepare_field_value(
                    field_name, field, value, is_complex
                )
                profile_settings[key] = prepared_value
        return profile_settings


DEFAULT_PREFECT_TOML_PATH = Path("prefect.toml")


class TomlConfigSettingsSourceBase(PydanticBaseSettingsSource, ConfigFileSourceMixin):
    def __init__(self, settings_cls: Type[BaseSettings]):
        super().__init__(settings_cls)
        self.settings_cls = settings_cls
        self.toml_data: dict[str, Any] = {}

    def _read_file(self, path: Path) -> dict[str, Any]:
        return _read_toml_file(path)

    @staticmethod
    def _field_is_dict_type(field: FieldInfo) -> bool:
        """Return True if the field's annotation is a dict type."""
        annotation = field.annotation
        if annotation is None:
            return False
        return get_origin(annotation) is dict or annotation is dict

    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> tuple[Any, str, bool]:
        """Concrete implementation to get the field value from toml data"""
        value = self.toml_data.get(field_name)
        if isinstance(value, dict) and not self._field_is_dict_type(field):
            # Dict values from TOML are likely nested settings objects and a
            # nested source will handle them — but not if the field itself
            # is typed as a dict (e.g. dict[str, ...]).
            value = None
        name = field_name
        # Use validation alias as the key to ensure profile value does not
        # higher priority sources. Lower priority sources that use the
        # field name can override higher priority sources that use the
        # validation alias as seen in https://github.com/PrefectHQ/prefect/issues/15981
        if value is not None:
            if field.validation_alias and isinstance(field.validation_alias, str):
                name = field.validation_alias
            elif field.validation_alias and isinstance(
                field.validation_alias, AliasChoices
            ):
                for alias in reversed(field.validation_alias.choices):
                    if isinstance(alias, str):
                        name = alias
                        break
        return value, name, self.field_is_complex(field)

    def prepare_field_value(
        self,
        field_name: str,
        field: FieldInfo,
        value: Any,
        value_is_complex: bool,
    ) -> Any:
        """Override to skip JSON decoding for dict values already parsed from TOML."""
        if isinstance(value, dict) and self._field_is_dict_type(field):
            return value
        return super().prepare_field_value(field_name, field, value, value_is_complex)

    def __call__(self) -> dict[str, Any]:
        """Called by pydantic to get the settings from our custom source"""
        toml_setings: dict[str, Any] = {}
        for field_name, field in self.settings_cls.model_fields.items():
            value, key, is_complex = self.get_field_value(field, field_name)
            if value is not None:
                prepared_value = self.prepare_field_value(
                    field_name, field, value, is_complex
                )
                toml_setings[key] = prepared_value
        return toml_setings


class PrefectTomlConfigSettingsSource(TomlConfigSettingsSourceBase):
    """Custom pydantic settings source to load settings from a prefect.toml file"""

    def __init__(
        self,
        settings_cls: Type[BaseSettings],
    ):
        super().__init__(settings_cls)
        self.toml_file_path: Path | str | Sequence[Path | str] | None = (
            settings_cls.model_config.get("toml_file", DEFAULT_PREFECT_TOML_PATH)
        )
        self.toml_data: dict[str, Any] = self._read_files(self.toml_file_path)
        self.toml_table_header: tuple[str, ...] = settings_cls.model_config.get(
            "prefect_toml_table_header", tuple()
        )
        for key in self.toml_table_header:
            self.toml_data: dict[str, Any] = self.toml_data.get(key, {})


class PyprojectTomlConfigSettingsSource(TomlConfigSettingsSourceBase):
    """Custom pydantic settings source to load settings from a pyproject.toml file"""

    def __init__(
        self,
        settings_cls: Type[BaseSettings],
    ):
        super().__init__(settings_cls)
        self.toml_file_path: Path = Path("pyproject.toml")
        self.toml_data: dict[str, Any] = self._read_files(self.toml_file_path)
        self.toml_table_header: tuple[str, ...] = settings_cls.model_config.get(
            "pyproject_toml_table_header", ("tool", "prefect")
        )
        for key in self.toml_table_header:
            self.toml_data: dict[str, Any] = self.toml_data.get(key, {})


def _is_test_mode() -> bool:
    """Check if the current process is in test mode."""
    return bool(
        os.getenv("PREFECT_TEST_MODE")
        or os.getenv("PREFECT_UNIT_TEST_MODE")
        or os.getenv("PREFECT_TESTING_UNIT_TEST_MODE")
        or os.getenv("PREFECT_TESTING_TEST_MODE")
    )


def _get_profiles_path() -> Path:
    """Helper to get the profiles path"""

    if _is_test_mode():
        return DEFAULT_PROFILES_PATH
    if env_path := os.getenv("PREFECT_PROFILES_PATH"):
        return Path(env_path)
    if dotenv_path := dotenv.dotenv_values(".env").get("PREFECT_PROFILES_PATH"):
        return Path(dotenv_path)
    if toml_path := _get_profiles_path_from_toml("prefect.toml", ["profiles_path"]):
        return Path(toml_path)
    if pyproject_path := _get_profiles_path_from_toml(
        "pyproject.toml", ["tool", "prefect", "profiles_path"]
    ):
        return Path(pyproject_path)

    if os.environ.get("PREFECT_HOME"):
        return Path(os.environ["PREFECT_HOME"]) / "profiles.toml"

    if not (DEFAULT_PREFECT_HOME / "profiles.toml").exists():
        return DEFAULT_PROFILES_PATH
    return DEFAULT_PREFECT_HOME / "profiles.toml"


def _get_profiles_path_from_toml(path: str, keys: List[str]) -> Optional[str]:
    """Helper to get the profiles path from a toml file."""

    try:
        toml_data = _read_toml_file(Path(path))
    except FileNotFoundError:
        return None

    return get_from_dict(toml_data, keys)



from typing import TYPE_CHECKING, overload
from prefect.cli.transfer._migratable_resources.base import (
    MigratableProtocol,
    MigratableType,
)


if TYPE_CHECKING:
    from prefect.client.schemas.objects import (
        BlockDocument,
        BlockSchema,
        BlockType,
        Flow,
        WorkPool,
        WorkQueue,
        Variable,
    )
    from prefect.client.schemas.responses import (
        DeploymentResponse,
        GlobalConcurrencyLimitResponse,
    )
    from prefect.events.schemas.automations import Automation

    from prefect.cli.transfer._migratable_resources.blocks import (
        MigratableBlockDocument,
        MigratableBlockSchema,
        MigratableBlockType,
    )
    from prefect.cli.transfer._migratable_resources.concurrency_limits import (
        MigratableGlobalConcurrencyLimit,
    )
    from prefect.cli.transfer._migratable_resources.deployments import (
        MigratableDeployment,
    )
    from prefect.cli.transfer._migratable_resources.flows import MigratableFlow
    from prefect.cli.transfer._migratable_resources.variables import MigratableVariable
    from prefect.cli.transfer._migratable_resources.work_pools import MigratableWorkPool
    from prefect.cli.transfer._migratable_resources.work_queues import (
        MigratableWorkQueue,
    )
    from prefect.cli.transfer._migratable_resources.automations import (
        MigratableAutomation,
    )


@overload
async def construct_migratable_resource(obj: "WorkPool") -> "MigratableWorkPool": ...
@overload
async def construct_migratable_resource(obj: "WorkQueue") -> "MigratableWorkQueue": ...
@overload
async def construct_migratable_resource(
    obj: "DeploymentResponse",
) -> "MigratableDeployment": ...
@overload
async def construct_migratable_resource(obj: "Flow") -> "MigratableFlow": ...
@overload
async def construct_migratable_resource(obj: "BlockType") -> "MigratableBlockType": ...
@overload
async def construct_migratable_resource(
    obj: "BlockSchema",
) -> "MigratableBlockSchema": ...
@overload
async def construct_migratable_resource(
    obj: "BlockDocument",
) -> "MigratableBlockDocument": ...
@overload
async def construct_migratable_resource(
    obj: "Automation",
) -> "MigratableAutomation": ...
@overload
async def construct_migratable_resource(
    obj: "GlobalConcurrencyLimitResponse",
) -> "MigratableGlobalConcurrencyLimit": ...
@overload
async def construct_migratable_resource(obj: "Variable") -> "MigratableVariable": ...


async def construct_migratable_resource(
    obj: MigratableType,
) -> MigratableProtocol:
    from prefect.client.schemas.objects import (
        BlockDocument,
        BlockSchema,
        BlockType,
        Flow,
        WorkPool,
        WorkQueue,
    )
    from prefect.client.schemas.responses import (
        DeploymentResponse,
        GlobalConcurrencyLimitResponse,
    )
    from prefect.events.schemas.automations import Automation

    from prefect.cli.transfer._migratable_resources.blocks import (
        MigratableBlockDocument,
        MigratableBlockSchema,
        MigratableBlockType,
    )
    from prefect.cli.transfer._migratable_resources.concurrency_limits import (
        MigratableGlobalConcurrencyLimit,
    )
    from prefect.cli.transfer._migratable_resources.deployments import (
        MigratableDeployment,
    )
    from prefect.cli.transfer._migratable_resources.flows import MigratableFlow
    from prefect.cli.transfer._migratable_resources.variables import MigratableVariable
    from prefect.cli.transfer._migratable_resources.work_pools import MigratableWorkPool
    from prefect.cli.transfer._migratable_resources.work_queues import (
        MigratableWorkQueue,
    )
    from prefect.cli.transfer._migratable_resources.automations import (
        MigratableAutomation,
    )

    if isinstance(obj, WorkPool):
        return await MigratableWorkPool.construct(obj)
    elif isinstance(obj, WorkQueue):
        return await MigratableWorkQueue.construct(obj)
    elif isinstance(obj, DeploymentResponse):
        return await MigratableDeployment.construct(obj)
    elif isinstance(obj, Flow):
        return await MigratableFlow.construct(obj)
    elif isinstance(obj, BlockType):
        return await MigratableBlockType.construct(obj)
    elif isinstance(obj, BlockSchema):
        return await MigratableBlockSchema.construct(obj)
    elif isinstance(obj, BlockDocument):
        return await MigratableBlockDocument.construct(obj)
    elif isinstance(obj, Automation):
        return await MigratableAutomation.construct(obj)
    elif isinstance(obj, GlobalConcurrencyLimitResponse):
        return await MigratableGlobalConcurrencyLimit.construct(obj)
    else:
        return await MigratableVariable.construct(obj)



import asyncio
from contextlib import asynccontextmanager
from logging import Logger
from typing import Any, AsyncGenerator, Callable

from docket import Docket, Worker

from prefect.logging import get_logger
from prefect.server.api.flow_runs import delete_flow_run_logs
from prefect.server.api.task_runs import delete_task_run_logs
from prefect.server.events.services import triggers as _triggers_module  # noqa: F401
from prefect.server.models.deployments import mark_deployments_ready
from prefect.server.models.work_queues import mark_work_queues_ready
from prefect.server.services.cancellation_cleanup import (
    cancel_child_task_runs,
    cancel_subflow_run,
)
from prefect.server.services.db_vacuum import (
    vacuum_events_with_retention_overrides,
    vacuum_old_events,
    vacuum_old_flow_runs,
    vacuum_orphaned_artifacts,
    vacuum_orphaned_logs,
    vacuum_stale_artifact_collections,
)
from prefect.server.services.late_runs import mark_flow_run_late
from prefect.server.services.pause_expirations import fail_expired_pause
from prefect.server.services.perpetual_services import (
    register_and_schedule_perpetual_services,
)
from prefect.server.services.repossessor import revoke_expired_lease

logger: Logger = get_logger(__name__)

# Task functions to register with docket for background processing
task_functions: list[Callable[..., Any]] = [
    # Simple background tasks (from Alex's PR #19377)
    mark_work_queues_ready,
    mark_deployments_ready,
    delete_task_run_logs,
    delete_flow_run_logs,
    # Find-and-flood pattern tasks used by perpetual services
    cancel_child_task_runs,
    cancel_subflow_run,
    fail_expired_pause,
    mark_flow_run_late,
    revoke_expired_lease,
    vacuum_orphaned_logs,
    vacuum_orphaned_artifacts,
    vacuum_stale_artifact_collections,
    vacuum_old_flow_runs,
    vacuum_events_with_retention_overrides,
    vacuum_old_events,
]


@asynccontextmanager
async def background_worker(
    docket: Docket,
    ephemeral: bool = False,
    webserver_only: bool = False,
) -> AsyncGenerator[None, None]:
    worker_task: asyncio.Task[None] | None = None
    async with Worker(docket) as worker:
        # Register background task functions
        docket.register_collection(
            "prefect.server.api.background_workers:task_functions"
        )

        # Register and schedule enabled perpetual services
        await register_and_schedule_perpetual_services(
            docket, ephemeral=ephemeral, webserver_only=webserver_only
        )

        try:
            worker_task = asyncio.create_task(worker.run_forever())
            yield

        finally:
            if worker_task:
                worker_task.cancel()
                try:
                    await worker_task
                except asyncio.CancelledError:
                    pass



"""
Internal implementation for SDK Analytics.

This module contains internal functions that should not be used directly.
Use the public API from prefect.analytics instead.
"""

from prefect._internal.analytics.emit import (
    _is_interactive_terminal,
    emit_integration_event,
    emit_sdk_event,
)
from prefect._internal.analytics.enabled import is_telemetry_enabled
from prefect._internal.analytics.milestones import (
    _mark_existing_user_milestones,
    try_mark_milestone,
)
from prefect._internal.analytics.notice import maybe_show_telemetry_notice

# Track initialization state
_telemetry_initialized = False


def initialize_analytics() -> None:
    """
    Initialize SDK analytics on Prefect import.

    This function:
    1. Checks if telemetry is enabled (quick local checks only - non-blocking)
    2. Detects existing users and pre-marks their milestones (no events emitted)
    3. For new users: shows the first-run notice and emits first_sdk_import

    Onboarding events are only emitted in interactive terminals to avoid
    tracking deployed flow runs (e.g., Kubernetes jobs) as new users.

    Called automatically when Prefect is imported.
    """
    global _telemetry_initialized

    if _telemetry_initialized:
        return

    _telemetry_initialized = True

    # Quick local checks only - no network calls
    if not is_telemetry_enabled():
        return

    # Only emit onboarding events in interactive terminals
    # This prevents deployed flow runs from being counted as new users
    if not _is_interactive_terminal():
        return

    try:
        # Check for existing users and pre-mark their milestones
        # This must happen BEFORE any events are emitted
        is_existing_user = _mark_existing_user_milestones()

        # Don't emit onboarding events for existing users
        if is_existing_user:
            return

        # Show first-run notice (only in interactive terminals)
        maybe_show_telemetry_notice()

        # Emit first_sdk_import event for new users only (tracked as a milestone
        # so it is only emitted once per installation)
        try_mark_milestone("first_sdk_import")
    except Exception:
        pass  # Silently ignore initialization errors


__all__ = [
    "initialize_analytics",
    "is_telemetry_enabled",
    "emit_sdk_event",
    "emit_integration_event",
    "try_mark_milestone",
    "_is_interactive_terminal",
]



"""
Automation command — native cyclopts implementation.

Manage automations.
"""

import asyncio
from pathlib import Path
from typing import Annotated, Optional
from uuid import UUID

import cyclopts

import prefect.cli._app as _cli
from prefect.cli._utilities import (
    exit_with_error,
    exit_with_success,
    with_cli_exception_handling,
)

automation_app: cyclopts.App = cyclopts.App(
    name="automation",
    alias="automations",
    help="Manage automations.",
    version_flags=[],
    help_flags=["--help"],
)


@automation_app.command(name="ls")
@with_cli_exception_handling
async def ls():
    """List all automations."""
    from rich.table import Table
    from rich.text import Text

    from prefect.client.orchestration import get_client

    async with get_client() as client:
        automations = await client.read_automations()

    table = Table(title="Automations", show_lines=True)

    table.add_column("Automation")
    table.add_column("Enabled")
    table.add_column("Trigger")
    table.add_column("Actions")

    for automation in automations:
        identifier_column = Text()

        identifier_column.append(automation.name, style="white bold")
        identifier_column.append("\n")

        identifier_column.append(str(automation.id), style="cyan")
        identifier_column.append("\n")

        if automation.description:
            identifier_column.append(automation.description, style="white")
            identifier_column.append("\n")

        if automation.actions_on_trigger or automation.actions_on_resolve:
            actions = (
                [
                    f"(trigger) {action.describe_for_cli()}"
                    for action in automation.actions
                ]
                + [
                    f"(trigger) {action.describe_for_cli()}"
                    for action in automation.actions_on_trigger
                ]
                + [
                    f"(resolve) {action.describe_for_cli()}"
                    for action in automation.actions
                ]
                + [
                    f"(resolve) {action.describe_for_cli()}"
                    for action in automation.actions_on_resolve
                ]
            )
        else:
            actions = [action.describe_for_cli() for action in automation.actions]

        table.add_row(
            identifier_column,
            str(automation.enabled),
            automation.trigger.describe_for_cli(),
            "\n".join(actions),
        )

    _cli.console.print(table)


@automation_app.command(name="inspect")
@with_cli_exception_handling
async def inspect(
    name: Annotated[Optional[str], cyclopts.Parameter(show=False)] = None,
    *,
    id: Annotated[
        Optional[str],
        cyclopts.Parameter("--id", help="An automation's id"),
    ] = None,
    yaml: Annotated[
        bool,
        cyclopts.Parameter("--yaml", help="Output as YAML"),
    ] = False,
    json: Annotated[
        bool,
        cyclopts.Parameter("--json", help="Output as JSON"),
    ] = False,
    output: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--output",
            alias="-o",
            help="Specify an output format. Currently supports: json, yaml",
        ),
    ] = None,
):
    """Inspect an automation."""
    import orjson
    import yaml as pyyaml
    from pydantic import BaseModel
    from rich.pretty import Pretty

    from prefect.client.orchestration import get_client
    from prefect.events.schemas.automations import Automation
    from prefect.exceptions import PrefectHTTPStatusError

    if output and output.lower() not in ["json", "yaml"]:
        exit_with_error("Only 'json' and 'yaml' output formats are supported.")

    if not id and not name:
        exit_with_error("Please provide either a name or an id.")

    automation: Automation | list[Automation] | None = None

    if name:
        async with get_client() as client:
            automation = await client.read_automations_by_name(name=name)
            if not automation:
                exit_with_error(f"Automation {name!r} not found.")

    elif id:
        async with get_client() as client:
            try:
                uuid_id = UUID(id)
                automation = await client.read_automation(uuid_id)
            except (PrefectHTTPStatusError, ValueError):
                exit_with_error(f"Automation with id {id!r} not found.")

    if yaml or json or (output and output.lower() in ["json", "yaml"]):

        def no_really_json(obj: type[BaseModel]):
            return orjson.loads(obj.model_dump_json())

        if isinstance(automation, list):
            automation = [no_really_json(a) for a in automation]
        elif isinstance(automation, Automation):
            automation = no_really_json(automation)

        if yaml or (output and output.lower() == "yaml"):
            _cli.console.print(pyyaml.dump(automation, sort_keys=False))
        elif json or (output and output.lower() == "json"):
            _cli.console.print(
                orjson.dumps(automation, option=orjson.OPT_INDENT_2).decode()
            )
    else:
        _cli.console.print(Pretty(automation))


@automation_app.command(name="resume", alias="enable")
@with_cli_exception_handling
async def resume(
    name: Annotated[Optional[str], cyclopts.Parameter(show=False)] = None,
    *,
    id: Annotated[
        Optional[str],
        cyclopts.Parameter("--id", help="An automation's id"),
    ] = None,
):
    """Resume an automation."""
    from prefect.client.orchestration import get_client
    from prefect.exceptions import PrefectHTTPStatusError

    if not id and not name:
        exit_with_error("Please provide either a name or an id.")

    if name:
        async with get_client() as client:
            automation = await client.read_automations_by_name(name=name)
            if not automation:
                exit_with_error(
                    f"Automation with name {name!r} not found. You can also specify an id with the `--id` flag."
                )
            if len(automation) > 1:
                from prefect.cli._prompts import confirm

                if not confirm(
                    f"Multiple automations found with name {name!r}. Do you want to resume all of them?",
                    default=False,
                    console=_cli.console,
                ):
                    exit_with_error("Resume aborted.")

            for a in automation:
                await client.resume_automation(a.id)
            exit_with_success(
                f"Resumed automation(s) with name {name!r} and id(s) {', '.join([repr(str(a.id)) for a in automation])}."
            )

    elif id:
        async with get_client() as client:
            try:
                uuid_id = UUID(id)
                automation = await client.read_automation(uuid_id)
            except (PrefectHTTPStatusError, ValueError):
                exit_with_error(f"Automation with id {id!r} not found.")
            await client.resume_automation(automation.id)
            exit_with_success(f"Resumed automation with id {str(automation.id)!r}.")


@automation_app.command(name="pause", alias="disable")
@with_cli_exception_handling
async def pause(
    name: Annotated[Optional[str], cyclopts.Parameter(show=False)] = None,
    *,
    id: Annotated[
        Optional[str],
        cyclopts.Parameter("--id", help="An automation's id"),
    ] = None,
):
    """Pause an automation."""
    from prefect.client.orchestration import get_client
    from prefect.exceptions import PrefectHTTPStatusError

    if not id and not name:
        exit_with_error("Please provide either a name or an id.")

    if name:
        async with get_client() as client:
            automation = await client.read_automations_by_name(name=name)
            if not automation:
                exit_with_error(
                    f"Automation with name {name!r} not found. You can also specify an id with the `--id` flag."
                )
            if len(automation) > 1:
                from prefect.cli._prompts import confirm

                if not confirm(
                    f"Multiple automations found with name {name!r}. Do you want to pause all of them?",
                    default=False,
                    console=_cli.console,
                ):
                    exit_with_error("Pause aborted.")

            for a in automation:
                await client.pause_automation(a.id)
            exit_with_success(
                f"Paused automation(s) with name {name!r} and id(s) {', '.join([repr(str(a.id)) for a in automation])}."
            )

    elif id:
        async with get_client() as client:
            try:
                uuid_id = UUID(id)
                automation = await client.read_automation(uuid_id)
            except (PrefectHTTPStatusError, ValueError):
                exit_with_error(f"Automation with id {id!r} not found.")
            await client.pause_automation(automation.id)
            exit_with_success(f"Paused automation with id {str(automation.id)!r}.")


@automation_app.command(name="delete")
@with_cli_exception_handling
async def delete(
    name: Annotated[Optional[str], cyclopts.Parameter(show=False)] = None,
    *,
    id: Annotated[
        Optional[str],
        cyclopts.Parameter("--id", help="An automation's id"),
    ] = None,
    all: Annotated[
        bool,
        cyclopts.Parameter("--all", help="Delete all automations"),
    ] = False,
):
    """Delete an automation."""
    from prefect.client.orchestration import get_client

    async with get_client() as client:
        if all:
            if name is not None or id is not None:
                exit_with_error(
                    "Cannot provide an automation name or id when deleting all automations."
                )
            automations = await client.read_automations()
            if len(automations) == 0:
                exit_with_success("No automations found.")

            if _cli.is_interactive():
                from prefect.cli._prompts import confirm

                if not confirm(
                    f"Are you sure you want to delete all {len(automations)} automations?",
                    default=False,
                    console=_cli.console,
                ):
                    exit_with_error("Deletion aborted.")

            semaphore = asyncio.Semaphore(10)

            async def limited_delete(automation_id):
                async with semaphore:
                    await client.delete_automation(automation_id)

            await asyncio.gather(
                *[limited_delete(automation.id) for automation in automations]
            )

            plural = "" if len(automations) == 1 else "s"
            exit_with_success(f"Deleted {len(automations)} automation{plural}.")

        if not id and not name:
            exit_with_error("Please provide either a name or an id.")

        if id:
            automation = await client.read_automation(id)
            if not automation:
                exit_with_error(f"Automation with id {id!r} not found.")

            if _cli.is_interactive():
                from prefect.cli._prompts import confirm

                if not confirm(
                    f"Are you sure you want to delete automation with id {id!r}?",
                    default=False,
                    console=_cli.console,
                ):
                    exit_with_error("Deletion aborted.")
            await client.delete_automation(id)
            exit_with_success(f"Deleted automation with id {id!r}")

        elif name:
            automation = await client.read_automations_by_name(name=name)
            if not automation:
                exit_with_error(
                    f"Automation {name!r} not found. You can also specify an id with the `--id` flag."
                )
            elif len(automation) > 1:
                exit_with_error(
                    f"Multiple automations found with name {name!r}. Please specify an id with the `--id` flag instead."
                )

            if _cli.is_interactive():
                from prefect.cli._prompts import confirm

                if not confirm(
                    f"Are you sure you want to delete automation with name {name!r}?",
                    default=False,
                    console=_cli.console,
                ):
                    exit_with_error("Deletion aborted.")
            await client.delete_automation(automation[0].id)
            exit_with_success(f"Deleted automation with name {name!r}")


@automation_app.command(name="create")
@with_cli_exception_handling
async def create(
    *,
    from_file: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--from-file",
            alias="-f",
            help="Path to YAML or JSON file containing automation(s)",
        ),
    ] = None,
    from_json: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--from-json",
            alias="-j",
            help="JSON string containing automation(s)",
        ),
    ] = None,
):
    """Create one or more automations from a file or JSON string."""
    import orjson
    import yaml as pyyaml

    from prefect.client.orchestration import get_client
    from prefect.events.schemas.automations import AutomationCore

    if from_file and from_json:
        exit_with_error("Please provide either --from-file or --from-json, not both.")

    if not from_file and not from_json:
        exit_with_error("Please provide either --from-file or --from-json.")

    if from_file:
        file_path = Path(from_file)
        if not file_path.exists():
            exit_with_error(f"File not found: {from_file}")

        with open(file_path, "r") as f:
            content = f.read()

        if file_path.suffix.lower() in [".yaml", ".yml"]:
            data = pyyaml.safe_load(content)
        elif file_path.suffix.lower() == ".json":
            data = orjson.loads(content)
        else:
            exit_with_error(
                "File extension not recognized. Please use .yaml, .yml, or .json"
            )
    else:  # from_json
        try:
            data = orjson.loads(from_json)
        except orjson.JSONDecodeError as e:
            exit_with_error(f"Invalid JSON: {e}")

    automations_data = []
    if isinstance(data, dict) and "automations" in data:
        automations_data = data["automations"]
    elif isinstance(data, list):
        automations_data = data
    else:
        automations_data = [data]

    created = []
    failed = []
    async with get_client() as client:
        for i, automation_data in enumerate(automations_data):
            try:
                automation = AutomationCore.model_validate(automation_data)
                automation_id = await client.create_automation(automation)
                created.append((automation.name, automation_id))
            except Exception as e:
                name = automation_data.get("name", f"automation at index {i}")
                failed.append((name, str(e)))

    if failed:
        _cli.console.print(f"[red]Failed to create {len(failed)} automation(s):[/red]")
        for name, error in failed:
            _cli.console.print(f"  - {name}: {error}")

    if created:
        if len(created) == 1 and not failed:
            name, automation_id = created[0]
            exit_with_success(f"Created automation '{name}' with id {automation_id}")
        else:
            _cli.console.print(f"[green]Created {len(created)} automation(s):[/green]")
            for name, automation_id in created:
                _cli.console.print(f"  - '{name}' with id {automation_id}")

    if failed:
        raise SystemExit(1)
    else:
        raise SystemExit(0)


@automation_app.command(name="update")
@with_cli_exception_handling
async def update(
    *,
    id: Annotated[
        str,
        cyclopts.Parameter("--id", help="The ID of the automation to update"),
    ],
    from_file: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--from-file",
            alias="-f",
            help="Path to YAML or JSON file containing the updated automation",
        ),
    ] = None,
    from_json: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--from-json",
            alias="-j",
            help="JSON string containing the updated automation",
        ),
    ] = None,
):
    """Update an existing automation from a file or JSON string."""
    import orjson
    import yaml as pyyaml

    from prefect.client.orchestration import get_client
    from prefect.events.schemas.automations import AutomationCore
    from prefect.exceptions import PrefectHTTPStatusError

    if from_file and from_json:
        exit_with_error("Please provide either --from-file or --from-json, not both.")

    if not from_file and not from_json:
        exit_with_error("Please provide either --from-file or --from-json.")

    try:
        automation_id = UUID(id)
    except ValueError:
        exit_with_error(f"Invalid automation ID: {id!r}")

    if from_file:
        file_path = Path(from_file)
        if not file_path.exists():
            exit_with_error(f"File not found: {from_file}")

        with open(file_path, "r") as f:
            content = f.read()

        if file_path.suffix.lower() in [".yaml", ".yml"]:
            data = pyyaml.safe_load(content)
        elif file_path.suffix.lower() == ".json":
            data = orjson.loads(content)
        else:
            exit_with_error(
                "File extension not recognized. Please use .yaml, .yml, or .json"
            )
    else:  # from_json
        try:
            data = orjson.loads(from_json)
        except orjson.JSONDecodeError as e:
            exit_with_error(f"Invalid JSON: {e}")

    try:
        automation = AutomationCore.model_validate(data)
    except Exception as e:
        exit_with_error(f"Automation config failed validation: {e}")

    async with get_client() as client:
        try:
            existing_automation = await client.read_automation(automation_id)
        except PrefectHTTPStatusError as e:
            if e.response.status_code == 404:
                exit_with_error(f"Automation with id {id!r} not found.")
            raise

        if not existing_automation:
            exit_with_error(f"Automation with id {id!r} not found.")

        await client.update_automation(automation_id, automation)
        exit_with_success(f"Updated automation '{automation.name}' ({id})")



"""Centralized registry of infrastructure process exit code explanations and
resolution hints.

Consumed by the worker and runner to emit actionable log messages when flow run
infrastructure exits with a non-zero status code.
"""

import logging
from dataclasses import dataclass


@dataclass(frozen=True)
class InfrastructureExitInfo:
    """Human-readable explanation and resolution for an infrastructure process exit code."""

    explanation: str
    resolution: str
    log_level: int = logging.ERROR


INFRASTRUCTURE_EXIT_HINTS: dict[int, InfrastructureExitInfo] = {
    0: InfrastructureExitInfo(
        explanation="Process exited cleanly.",
        resolution="No action needed.",
        log_level=logging.INFO,
    ),
    -9: InfrastructureExitInfo(
        explanation=(
            "Process exited due to a SIGKILL signal. Typically caused by the"
            " operating system terminating the process for exceeding memory limits,"
            " or by manual cancellation."
        ),
        resolution=(
            "Check whether the flow run exceeded its memory allocation. If running"
            " in a container, increase the memory limit. If this was intentional"
            " cancellation, no action is needed."
        ),
        log_level=logging.INFO,
    ),
    -15: InfrastructureExitInfo(
        explanation=(
            "Process exited due to a SIGTERM signal. Typically caused by graceful"
            " shutdown or manual cancellation."
        ),
        resolution=(
            "If this was caused by cancellation, no action is needed. Otherwise,"
            " check for infrastructure scaling events or deployment rollovers."
        ),
        log_level=logging.INFO,
    ),
    1: InfrastructureExitInfo(
        explanation="Process exited with a general error.",
        resolution=(
            "Check the flow run logs for an unhandled exception or assertion error."
        ),
    ),
    125: InfrastructureExitInfo(
        explanation="Container failed to run (Docker/OCI exit code 125).",
        resolution=(
            "Verify the container image exists, the entrypoint is correct, and"
            " the container runtime is healthy."
        ),
    ),
    126: InfrastructureExitInfo(
        explanation="Command found but not executable (permission denied).",
        resolution=(
            "Check file permissions on the entrypoint script or binary. Ensure the"
            " file has the executable bit set."
        ),
    ),
    127: InfrastructureExitInfo(
        explanation="Command not found.",
        resolution=(
            "Verify that the entrypoint command or script is installed in the"
            " execution environment. Check your PATH and container image contents."
        ),
    ),
    137: InfrastructureExitInfo(
        explanation="Process was killed, likely due to an out-of-memory (OOM) condition.",
        resolution=(
            "Increase the memory allocation for the flow run infrastructure."
            " In Kubernetes, raise the memory limit in the job template."
            " In ECS, increase the task memory."
        ),
    ),
    143: InfrastructureExitInfo(
        explanation=(
            "Process received SIGTERM via the container runtime. This is the"
            " containerized equivalent of exit code -15."
        ),
        resolution=(
            "If this was caused by cancellation or a scaling event, no action is"
            " needed. Otherwise, check for pod evictions or ECS task stops."
        ),
        log_level=logging.INFO,
    ),
    247: InfrastructureExitInfo(
        explanation="Process was terminated due to high memory usage.",
        resolution=(
            "Increase the memory allocation for the flow run infrastructure."
            " Consider profiling memory usage to find the source of consumption."
        ),
    ),
    # Windows Ctrl+C / Ctrl+Break (STATUS_CONTROL_C_EXIT)
    0xC000013A: InfrastructureExitInfo(
        explanation=(
            "Process was terminated due to a Ctrl+C or Ctrl+Break signal."
            " Typically caused by manual cancellation."
        ),
        resolution="If this was intentional cancellation, no action is needed.",
        log_level=logging.INFO,
    ),
}


def get_infrastructure_exit_info(code: int) -> InfrastructureExitInfo:
    """Return the `InfrastructureExitInfo` for *code*.

    Known codes return a specific explanation and resolution. Unknown non-zero
    codes return a generic entry. Code 0 always returns the "exited cleanly"
    entry.
    """
    try:
        return INFRASTRUCTURE_EXIT_HINTS[code]
    except KeyError:
        return InfrastructureExitInfo(
            explanation="Process exited with an unexpected status code.",
            resolution=(
                "Check the flow run logs and infrastructure logs for more details."
            ),
        )



class TransferSkipped(Exception):
    """
    Exception raised when a resource is skipped during transfer.
    """

    def __init__(self, reason: str):
        self.reason = reason

    def __str__(self) -> str:
        return self.reason



"""Manage Prefect Cloud assets."""

from __future__ import annotations

from typing import Annotated

import cyclopts
from rich.table import Table

import prefect.cli._app as _cli
from prefect.cli._cloud_utils import confirm_logged_in
from prefect.cli._utilities import (
    exit_with_error,
    exit_with_success,
    with_cli_exception_handling,
)

asset_app: cyclopts.App = cyclopts.App(
    name="asset",
    alias="assets",
    help="Manage Prefect Cloud assets.",
    version_flags=[],
    help_flags=["--help"],
)


@asset_app.command(name="ls")
@with_cli_exception_handling
async def asset_ls(
    *,
    prefix: Annotated[
        str | None,
        cyclopts.Parameter("--prefix", alias="-p", help="Filter assets by key prefix"),
    ] = None,
    search: Annotated[
        str | None,
        cyclopts.Parameter(
            "--search", alias="-s", help="Filter assets by key substring"
        ),
    ] = None,
    limit: Annotated[
        int,
        cyclopts.Parameter(
            "--limit",
            alias="-l",
            help="Maximum number of assets to return (default 50, max 200)",
        ),
    ] = 50,
    output: Annotated[
        str | None,
        cyclopts.Parameter(
            "--output", alias="-o", help="Output format. Supports: json"
        ),
    ] = None,
):
    """List assets in the current workspace."""
    import orjson

    from prefect.client.cloud import get_cloud_client
    from prefect.settings import get_current_settings

    confirm_logged_in(console=_cli.console)

    if output and output.lower() != "json":
        exit_with_error("Only 'json' output format is supported.")

    if limit < 1 or limit > 200:
        exit_with_error("Limit must be between 1 and 200.")

    key_filter: dict[str, list[str]] = {}
    if prefix:
        key_filter["prefix"] = [prefix]
    if search:
        key_filter["search"] = [search]

    body: dict[str, object] = {"limit": limit}
    if key_filter:
        body["filter"] = {"key": key_filter}

    async with get_cloud_client(host=get_current_settings().api.url) as client:
        response = await client.request("POST", "/assets/filter", json=body)

    assets = response.get("assets", [])
    total = response.get("total", len(assets))

    if output and output.lower() == "json":
        json_output = orjson.dumps(assets, option=orjson.OPT_INDENT_2).decode()
        _cli.console.print(json_output, soft_wrap=True)
    else:
        if not assets:
            _cli.console.print("No assets found in this workspace.")
            return

        table = Table(
            title="Assets",
            show_header=True,
        )

        table.add_column("Key", style="blue", no_wrap=False)
        table.add_column("Last Seen", style="cyan", no_wrap=True)

        for asset in sorted(assets, key=lambda x: x.get("key", "")):
            table.add_row(
                asset.get("key", ""),
                asset.get("last_seen", ""),
            )

        _cli.console.print(table)
        _cli.console.print(f"\nShowing {len(assets)} of {total} asset(s)")


@asset_app.command(name="delete")
@with_cli_exception_handling
async def asset_delete(
    key: Annotated[str, cyclopts.Parameter(help="The key of the asset to delete")],
    *,
    force: Annotated[
        bool,
        cyclopts.Parameter(
            "--force", alias="-f", negative="", help="Skip confirmation prompt"
        ),
    ] = False,
):
    """Delete an asset by its key.

    The key should be the full asset URI (e.g., 's3://bucket/data.csv').
    """
    from prefect.cli._prompts import confirm
    from prefect.client.cloud import get_cloud_client
    from prefect.exceptions import ObjectNotFound
    from prefect.settings import get_current_settings

    confirm_logged_in(console=_cli.console)

    if _cli.is_interactive() and not force:
        if not confirm(
            f"Are you sure you want to delete asset {key!r}?",
            default=False,
            console=_cli.console,
        ):
            exit_with_error("Deletion aborted.")

    async with get_cloud_client(host=get_current_settings().api.url) as client:
        try:
            await client.request("DELETE", "/assets/key", params={"key": key})
        except ObjectNotFound:
            exit_with_error(f"Asset {key!r} not found.")

    exit_with_success(f"Deleted asset {key!r}.")



from __future__ import annotations

from pathlib import Path
from typing import Any


class _PullStepStorage:
    """
    A shim storage class that allows passing pull steps to a `RunnerDeployment`.
    """

    def __init__(self, pull_steps: list[dict[str, Any]]):
        self._base_path = Path.cwd()
        self.pull_steps = pull_steps

    def set_base_path(self, path: Path):
        self._base_path = path

    @property
    def destination(self):
        return self._base_path

    @property
    def pull_interval(self):
        return 60

    async def pull_code(self):
        pass

    def to_pull_step(self):
        return self.pull_steps

    def __eq__(self, other: Any) -> bool:
        return self.pull_steps == getattr(other, "pull_steps", None)



"""Functionality we want to use from later Python versions."""

try:
    import tomllib  # 3.11+
except ImportError:
    import toml as tomllib  # fallback on Python <3.11

    tomllib.TOMLDecodeError = tomllib.TomlDecodeError


__all__ = ["tomllib"]



"""
The Repossessor service. Handles reconciliation of expired concurrency leases.
"""

from __future__ import annotations

import logging
from datetime import datetime, timedelta, timezone
from typing import Annotated
from uuid import UUID

from docket import CurrentDocket, Depends, Docket, Logged, Perpetual

from prefect.logging import get_logger
from prefect.server.concurrency.lease_storage import (
    ConcurrencyLeaseStorage,
    get_concurrency_lease_storage,
)
from prefect.server.database import PrefectDBInterface, provide_database_interface
from prefect.server.models.concurrency_limits_v2 import bulk_decrement_active_slots
from prefect.server.services.perpetual_services import perpetual_service
from prefect.settings.context import get_current_settings

logger: logging.Logger = get_logger(__name__)


async def revoke_expired_lease(
    lease_id: Annotated[UUID, Logged],
    *,
    db: PrefectDBInterface = Depends(provide_database_interface),
    lease_storage: ConcurrencyLeaseStorage = Depends(get_concurrency_lease_storage),
) -> None:
    """Revoke a single expired lease (docket task)."""
    expired_lease = await lease_storage.read_lease(lease_id)
    if expired_lease is None or expired_lease.metadata is None:
        logger.warning(
            f"Lease {lease_id} should be revoked but was not found or has no metadata"
        )
        return

    occupancy_seconds = (
        datetime.now(timezone.utc) - expired_lease.created_at
    ).total_seconds()

    logger.info(
        f"Revoking lease {lease_id} for {len(expired_lease.resource_ids)} "
        f"concurrency limits with {expired_lease.metadata.slots} slots"
    )

    async with db.session_context(begin_transaction=True) as session:
        await bulk_decrement_active_slots(
            session=session,
            concurrency_limit_ids=expired_lease.resource_ids,
            slots=expired_lease.metadata.slots,
            occupancy_seconds=occupancy_seconds,
        )
        await lease_storage.revoke_lease(lease_id)


@perpetual_service(
    enabled_getter=lambda: get_current_settings().server.services.repossessor.enabled,
)
async def monitor_expired_leases(
    docket: Docket = CurrentDocket(),
    lease_storage: ConcurrencyLeaseStorage = Depends(get_concurrency_lease_storage),
    perpetual: Perpetual = Perpetual(
        automatic=False,
        every=timedelta(
            seconds=get_current_settings().server.services.repossessor.loop_seconds
        ),
    ),
) -> None:
    """Monitor for expired leases and schedule revocation tasks."""
    expired_lease_ids = await lease_storage.read_expired_lease_ids()

    if expired_lease_ids:
        logger.info(f"Scheduling revocation of {len(expired_lease_ids)} expired leases")

    for lease_id in expired_lease_ids:
        await docket.add(revoke_expired_lease)(lease_id)



"""Add expression index for per-schedule scheduler query

Revision ID: a1b2c3d4e5f6
Revises: 9e83011d1f2a
Create Date: 2026-02-19 20:00:00.000000

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "a1b2c3d4e5f6"
down_revision = "9e83011d1f2a"
branch_labels = None
depends_on = None


def upgrade():
    with op.get_context().autocommit_block():
        op.execute(
            """
            CREATE INDEX CONCURRENTLY IF NOT EXISTS
            ix_flow_run__schedule_id_scheduler
            ON flow_run (deployment_id, (created_by->>'id'), next_scheduled_start_time)
            WHERE state_type = 'SCHEDULED' AND auto_scheduled = true;
            """
        )


def downgrade():
    with op.get_context().autocommit_block():
        op.execute(
            """
            DROP INDEX CONCURRENTLY IF EXISTS ix_flow_run__schedule_id_scheduler;
            """
        )



"""Add composite index on task_run(state_type, start_time)

Revision ID: 09a9e091e578
Revises: a1b2c3d4e5f6
Create Date: 2026-03-03 00:00:00.000000

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "09a9e091e578"
down_revision = "a1b2c3d4e5f6"
branch_labels = None
depends_on = None


def upgrade():
    with op.get_context().autocommit_block():
        op.execute(
            """
            CREATE INDEX CONCURRENTLY IF NOT EXISTS
            ix_task_run__state_type_start_time
            ON task_run (state_type, start_time)
            """
        )


def downgrade():
    with op.get_context().autocommit_block():
        op.execute(
            "DROP INDEX CONCURRENTLY IF EXISTS ix_task_run__state_type_start_time"
        )



"""
Experimental plugin API specification.

This module defines the hook specification and data structures for Prefect's
experimental plugin system.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any, Callable, Mapping, Optional

import pluggy

# Bump this when breaking the hook contract
PREFECT_PLUGIN_API_VERSION = "0.1"

hookspec = pluggy.HookspecMarker("prefect-experimental")


@dataclass
class HookContext:
    """
    Context provided to plugin hooks at startup.

    Attributes:
        prefect_version: The version of Prefect running
        api_url: The configured Prefect API URL, if any
        logger_factory: Factory function to create a stdlib logger for the plugin
    """

    prefect_version: str
    api_url: str | None
    # Logger factory returns a stdlib logger; plugins should use this.
    logger_factory: Callable[[str], logging.Logger]
    # Future: async Prefect client getter, settings snapshot, etc.


@dataclass
class SetupResult:
    """
    Result returned by a plugin's setup_environment hook.

    Attributes:
        env: Environment variables to set (e.g., AWS_* variables)
        note: Short, non-secret human-readable hint about what was configured
        required: If True and hook fails, abort in strict mode
    """

    env: Mapping[str, str]  # e.g. AWS_* variables
    note: Optional[str] = None  # short, non-secret human hint
    required: bool = False  # if True and hook fails -> abort (in strict mode)


class HookSpec:
    """
    Plugin hook specification.

    Plugins should implement the methods defined here to provide startup hooks.
    """

    @hookspec
    def setup_environment(self, *, ctx: HookContext) -> Optional[SetupResult]:
        """
        Prepare process environment for Prefect and its children.

        This hook is called before Prefect CLI commands, workers, or agents
        start their main work. It allows plugins to configure environment
        variables, authenticate with external services, or perform other
        setup tasks.

        Args:
            ctx: Context object with Prefect version, API URL, and logger factory

        Returns:
            SetupResult with environment variables to set, or None to indicate
            no changes are needed.

        Important:
            - Must not print secrets or write sensitive data to disk by default
            - Should be idempotent
            - May be async or sync
            - Exceptions are caught and logged unless required=True in strict mode
        """

    @hookspec
    def set_database_connection_params(
        self, connection_url: str, settings: Any
    ) -> Mapping[str, Any]:
        """
        Set additional database connection parameters.

        This hook is called when creating a database engine. It allows plugins
        to provide additional connection parameters, such as authentication
        tokens or SSL configuration.

        Args:
            connection_url: The database connection URL
            settings: The current Prefect settings

        Returns:
            Dictionary of connection parameters to merge into connect_args
        """



"""
Experimental command — native cyclopts implementation.

Access experimental features (subject to change).
"""

import cyclopts

import prefect.cli._app as _cli
from prefect.cli._utilities import with_cli_exception_handling

experimental_app: cyclopts.App = cyclopts.App(
    name="experimental",
    help="Access experimental features (subject to change).",
    version_flags=[],
    help_flags=["--help"],
)

plugins_app: cyclopts.App = cyclopts.App(
    name="plugins",
    help="Plugin system diagnostics.",
    version_flags=[],
    help_flags=["--help"],
)
experimental_app.command(plugins_app)


@plugins_app.command(name="diagnose")
@with_cli_exception_handling
async def diagnose():
    """Diagnose the experimental plugin system."""
    import importlib.metadata as md

    from prefect._experimental.plugins.manager import ENTRYPOINTS_GROUP
    from prefect.settings import get_current_settings

    _cli.console.print(
        "\n[bold]Prefect Experimental Plugin System Diagnostics[/bold]\n"
    )

    # Check if enabled
    settings = get_current_settings().experiments.plugins
    enabled = settings.enabled
    _cli.console.print(f"Enabled: [{'green' if enabled else 'red'}]{enabled}[/]")

    if not enabled:
        _cli.console.print("\n[yellow]Plugin system is disabled.[/yellow]")
        _cli.console.print(
            "Set [cyan]PREFECT_EXPERIMENTS_PLUGINS_ENABLED=1[/cyan] to enable.\n"
        )
        return

    # Show configuration
    timeout = settings.setup_timeout_seconds
    allow = settings.allow
    deny = settings.deny
    strict = settings.strict
    safe = settings.safe_mode

    _cli.console.print(f"Timeout: {timeout}s")
    _cli.console.print(f"Strict mode: {strict}")
    _cli.console.print(f"Safe mode: {safe}")
    _cli.console.print(f"Allow list: {allow or 'None'}")
    _cli.console.print(f"Deny list: {deny or 'None'}")

    # Discover entry points
    _cli.console.print(
        f"\n[bold]Discoverable Plugins (entry point group: {ENTRYPOINTS_GROUP})[/bold]\n"
    )

    entry_points = list(md.entry_points(group=ENTRYPOINTS_GROUP))

    if not entry_points:
        _cli.console.print("[yellow]No plugins found.[/yellow]\n")
        return

    for ep in entry_points:
        filtered = False
        reason = None
        if allow and ep.name not in allow:
            filtered = True
            reason = "not in allow list"
        elif deny and ep.name in deny:
            filtered = True
            reason = "in deny list"

        status_color = "red" if filtered else "green"
        status = f"[{status_color}]{'filtered' if filtered else 'active'}[/]"

        _cli.console.print(f"  • {ep.name}: {status}")
        _cli.console.print(f"    Module: {ep.value}")

        if filtered and reason:
            _cli.console.print(f"    Reason: {reason}")

        # Try to load and get version requirement
        try:
            plugin = ep.load()
            requires = getattr(plugin, "PREFECT_PLUGIN_API_REQUIRES", ">=0.1,<1")
            _cli.console.print(f"    API requirement: {requires}")
        except Exception as e:
            _cli.console.print(f"    [red]Failed to load: {e}[/]")

        _cli.console.print()

    # Run startup hooks to show what they do
    if not safe:
        _cli.console.print("[bold]Running Startup Hooks[/bold]\n")

        from prefect import __version__
        from prefect._experimental.plugins import run_startup_hooks
        from prefect._experimental.plugins.spec import HookContext
        from prefect.logging import get_logger

        ctx = HookContext(
            prefect_version=__version__,
            api_url=get_current_settings().api.url,
            logger_factory=get_logger,
        )
        summaries = await run_startup_hooks(ctx)

        if summaries:
            for summary in summaries:
                status = "[red]error[/]" if summary.error else "[green]success[/]"
                _cli.console.print(f"  • {summary.plugin}: {status}")

                if summary.error:
                    _cli.console.print(f"    Error: {summary.error}")
                elif summary.env_preview:
                    _cli.console.print(
                        f"    Environment variables: {len(summary.env_preview)}"
                    )
                    for k, v in summary.env_preview.items():
                        _cli.console.print(f"      {k}={v}")

                    if summary.note:
                        _cli.console.print(f"    Note: {summary.note}")
                else:
                    _cli.console.print("    No changes")

                _cli.console.print()
        else:
            _cli.console.print("[yellow]No plugins executed.[/yellow]\n")
    else:
        _cli.console.print(
            "\n[yellow]Safe mode enabled - skipping hook execution.[/yellow]"
        )
        _cli.console.print(
            "Set [cyan]PREFECT_EXPERIMENTS_PLUGINS_SAFE_MODE=0[/cyan] to execute hooks.\n"
        )



"""
The CancellationCleanup service. Responsible for cancelling tasks and subflows that haven't finished.
"""

import datetime
from typing import Annotated
from uuid import UUID

import sqlalchemy as sa
from docket import CurrentDocket, Depends, Docket, Logged, Perpetual
from sqlalchemy.sql.expression import or_

import prefect.server.models as models
from prefect.server.database import PrefectDBInterface, provide_database_interface
from prefect.server.schemas import filters, states
from prefect.server.services.perpetual_services import perpetual_service
from prefect.settings.context import get_current_settings
from prefect.types._datetime import now

NON_TERMINAL_STATES = list(set(states.StateType) - states.TERMINAL_STATES)


# Docket task function for cancelling child task runs of a cancelled flow run
async def cancel_child_task_runs(
    flow_run_id: Annotated[UUID, Logged],
    *,
    db: PrefectDBInterface = Depends(provide_database_interface),
) -> None:
    """Cancel child task runs of a cancelled flow run (docket task)."""
    async with db.session_context() as session:
        child_task_runs = await models.task_runs.read_task_runs(
            session,
            flow_run_filter=filters.FlowRunFilter(
                id=filters.FlowRunFilterId(any_=[flow_run_id])
            ),
            task_run_filter=filters.TaskRunFilter(
                state=filters.TaskRunFilterState(
                    type=filters.TaskRunFilterStateType(any_=NON_TERMINAL_STATES)
                )
            ),
            limit=100,
        )

    for task_run in child_task_runs:
        async with db.session_context(begin_transaction=True) as session:
            await models.task_runs.set_task_run_state(
                session=session,
                task_run_id=task_run.id,
                state=states.Cancelled(message="The parent flow run was cancelled."),
                force=True,
            )


# Docket task function for cancelling a subflow run whose parent was cancelled
async def cancel_subflow_run(
    subflow_run_id: Annotated[UUID, Logged],
    *,
    db: PrefectDBInterface = Depends(provide_database_interface),
) -> None:
    """Cancel a subflow run whose parent flow run was cancelled (docket task)."""
    async with db.session_context() as session:
        flow_run = await models.flow_runs.read_flow_run(
            session, flow_run_id=subflow_run_id
        )

        if not flow_run or not flow_run.parent_task_run_id or not flow_run.state:
            return

        if flow_run.state.type in states.TERMINAL_STATES:
            return

        parent_task_run = await models.task_runs.read_task_run(
            session, task_run_id=flow_run.parent_task_run_id
        )

        if not parent_task_run or not parent_task_run.flow_run_id:
            return

        containing_flow_run = await models.flow_runs.read_flow_run(
            session, flow_run_id=parent_task_run.flow_run_id
        )

        if (
            containing_flow_run
            and containing_flow_run.state
            and containing_flow_run.state.type != states.StateType.CANCELLED
        ):
            return

        if flow_run.deployment_id:
            state = states.Cancelling(message="The parent flow run was cancelled.")
        else:
            state = states.Cancelled(message="The parent flow run was cancelled.")

    async with db.session_context(begin_transaction=True) as session:
        await models.flow_runs.set_flow_run_state(
            session=session,
            flow_run_id=subflow_run_id,
            state=state,
        )


# Perpetual monitor for cancelled flow runs with child tasks (find and flood pattern)
@perpetual_service(
    enabled_getter=lambda: (
        get_current_settings().server.services.cancellation_cleanup.enabled
    ),
)
async def monitor_cancelled_flow_runs(
    docket: Docket = CurrentDocket(),
    db: PrefectDBInterface = Depends(provide_database_interface),
    perpetual: Perpetual = Perpetual(
        automatic=False,
        every=datetime.timedelta(
            seconds=get_current_settings().server.services.cancellation_cleanup.loop_seconds
        ),
    ),
) -> None:
    """Monitor for cancelled flow runs and schedule child task cancellation."""

    batch_size = 200
    cancelled_flow_query = (
        sa.select(db.FlowRun.id)
        .where(
            db.FlowRun.state_type == states.StateType.CANCELLED,
            db.FlowRun.end_time.is_not(None),
            db.FlowRun.end_time >= (now("UTC") - datetime.timedelta(days=1)),
        )
        .order_by(db.FlowRun.id)
        .limit(batch_size)
    )

    async with db.session_context() as session:
        flow_run_result = await session.execute(cancelled_flow_query)
    flow_run_ids = flow_run_result.scalars().all()

    for flow_run_id in flow_run_ids:
        await docket.add(cancel_child_task_runs)(flow_run_id)


# Perpetual monitor for subflow runs that need cancellation (find and flood pattern)
@perpetual_service(
    enabled_getter=lambda: (
        get_current_settings().server.services.cancellation_cleanup.enabled
    ),
)
async def monitor_subflow_runs(
    docket: Docket = CurrentDocket(),
    db: PrefectDBInterface = Depends(provide_database_interface),
    perpetual: Perpetual = Perpetual(
        automatic=False,
        every=datetime.timedelta(
            seconds=get_current_settings().server.services.cancellation_cleanup.loop_seconds
        ),
    ),
) -> None:
    """Monitor for subflow runs that need to be cancelled."""

    batch_size = 200
    subflow_query = (
        sa.select(db.FlowRun.id)
        .where(
            or_(
                db.FlowRun.state_type == states.StateType.PENDING,
                db.FlowRun.state_type == states.StateType.SCHEDULED,
                db.FlowRun.state_type == states.StateType.RUNNING,
                db.FlowRun.state_type == states.StateType.PAUSED,
                db.FlowRun.state_type == states.StateType.CANCELLING,
            ),
            db.FlowRun.parent_task_run_id.is_not(None),
        )
        .order_by(db.FlowRun.id)
        .limit(batch_size)
    )

    async with db.session_context() as session:
        subflow_run_result = await session.execute(subflow_query)
    subflow_run_ids = subflow_run_result.scalars().all()

    for subflow_run_id in subflow_run_ids:
        await docket.add(cancel_subflow_run)(subflow_run_id)



from __future__ import annotations

import uuid

from typing_extensions import Self

from prefect.cli.transfer._exceptions import TransferSkipped
from prefect.cli.transfer._migratable_resources.base import (
    MigratableProtocol,
    MigratableResource,
)
from prefect.client.orchestration import get_client
from prefect.client.schemas.actions import (
    VariableCreate,
)
from prefect.client.schemas.objects import (
    Variable,
)
from prefect.exceptions import (
    ObjectAlreadyExists,
)


class MigratableVariable(MigratableResource[Variable]):
    _instances: dict[uuid.UUID, Self] = {}

    def __init__(self, variable: Variable):
        self.source_variable = variable
        self.destination_variable: Variable | None = None

    @property
    def source_id(self) -> uuid.UUID:
        return self.source_variable.id

    @property
    def destination_id(self) -> uuid.UUID | None:
        return self.destination_variable.id if self.destination_variable else None

    @classmethod
    async def construct(cls, obj: Variable) -> Self:
        if obj.id in cls._instances:
            return cls._instances[obj.id]
        instance = cls(obj)
        cls._instances[obj.id] = instance
        return instance

    @classmethod
    async def get_instance(cls, id: uuid.UUID) -> "MigratableResource[Variable] | None":
        if id in cls._instances:
            return cls._instances[id]
        return None

    async def get_dependencies(self) -> "list[MigratableProtocol]":
        return []

    async def migrate(self) -> None:
        async with get_client() as client:
            try:
                self.destination_variable = await client.create_variable(
                    variable=VariableCreate(
                        name=self.source_variable.name,
                        value=self.source_variable.value,
                        tags=self.source_variable.tags,
                    ),
                )
            except ObjectAlreadyExists:
                self.destination_variable = await client.read_variable_by_name(
                    self.source_variable.name
                )
                raise TransferSkipped("Already exists")



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

from pydantic import SecretStr, ValidationInfo

if TYPE_CHECKING:
    from prefect.settings.models.root import Settings


def default_profiles_path(values: dict[str, Any]) -> Path:
    """Default profiles_path based on home directory."""
    home = values.get("home", Path("~/.prefect").expanduser())
    if not isinstance(home, Path):
        home = Path("~/.prefect").expanduser()
    return home / "profiles.toml"


def substitute_home_template(v: Any, info: ValidationInfo) -> Any:
    """Validator that substitutes $PREFECT_HOME in a path string if present."""
    home_path = info.data.get("home")

    path_str: str | None = None
    if isinstance(v, Path):
        return v
    elif isinstance(v, str):
        path_str = v
    elif v is None:
        return None
    else:
        return v

    if path_str and "$PREFECT_HOME" in path_str:
        if home_path and isinstance(home_path, Path):
            resolved_str = path_str.replace("$PREFECT_HOME", str(home_path))
            try:
                return Path(resolved_str)
            except Exception as e:
                raise ValueError(
                    f"Error creating path after substituting $PREFECT_HOME: {e}"
                ) from e
        else:
            raise ValueError(
                f'Cannot resolve $PREFECT_HOME in "{path_str}" because '
                f"PREFECT_HOME setting ({home_path!r}) is not a valid resolved path."
            )

    return path_str


def default_local_storage_path(values: dict[str, Any]) -> Path:
    """Default local_storage_path based on home directory."""
    home = values.get("home")
    if not isinstance(home, Path):
        home = Path("~/.prefect").expanduser()
    return home / "storage"


def default_memo_store_path(values: dict[str, Any]) -> Path:
    """Default memo_store_path based on home directory."""
    home = values.get("home")
    if not isinstance(home, Path):
        home = Path("~/.prefect").expanduser()
    return home / "memo_store.toml"


def default_logging_config_path(values: dict[str, Any]) -> Path:
    """Default logging_config_path based on home directory."""
    home = values.get("home")
    if not isinstance(home, Path):
        home = Path("~/.prefect").expanduser()
    return home / "logging.yml"


def default_database_connection_url(settings: "Settings") -> SecretStr:
    value: str = f"sqlite+aiosqlite:///{settings.home}/prefect.db"
    if settings.server.database.driver == "postgresql+asyncpg":
        required = [
            "host",
            "user",
            "name",
        ]
        missing = [
            attr for attr in required if getattr(settings.server.database, attr) is None
        ]
        if missing:
            raise ValueError(
                f"Missing required database connection settings: {', '.join(missing)}"
            )

        from sqlalchemy import URL

        value = URL(
            drivername=settings.server.database.driver,
            host=settings.server.database.host,
            port=settings.server.database.port or 5432,
            username=settings.server.database.user,
            password=(
                settings.server.database.password.get_secret_value()
                if settings.server.database.password
                else None
            ),
            database=settings.server.database.name,
            query=[],  # type: ignore
        ).render_as_string(hide_password=False)

    elif settings.server.database.driver == "sqlite+aiosqlite":
        if settings.server.database.name:
            value = (
                f"{settings.server.database.driver}:///{settings.server.database.name}"
            )
        else:
            value = f"sqlite+aiosqlite:///{settings.home}/prefect.db"

    elif settings.server.database.driver:
        raise ValueError(
            f"Unsupported database driver: {settings.server.database.driver}"
        )
    return SecretStr(value)


def default_ui_url(settings: "Settings") -> str | None:
    value = settings.ui_url
    if value is not None:
        return value

    # Otherwise, infer a value from the API URL
    ui_url = api_url = settings.api.url

    if not api_url:
        return None
    assert ui_url is not None

    cloud_url = settings.cloud.api_url
    cloud_ui_url = settings.cloud.ui_url
    if api_url.startswith(cloud_url) and cloud_ui_url:
        ui_url = ui_url.replace(cloud_url, cloud_ui_url)

    if ui_url.endswith("/api"):
        # Handles open-source APIs
        ui_url = ui_url[:-4]

    # Handles Cloud APIs with content after `/api`
    ui_url = ui_url.replace("/api/", "/")

    # Update routing
    ui_url = ui_url.replace("/accounts/", "/account/")
    ui_url = ui_url.replace("/workspaces/", "/workspace/")

    return ui_url
